/* Continuous Time Bayesian Network Reasoning and Learning Engine
 * Copyright (C) 2009 The Regents of the University of California
 *
 * Authored by Yu Fan, William Lam, Joon Lee, Christian Shelton, and Jing Xu
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef _BN_H_
#define _BN_H_

#include "rv.h"
#include "structure.h"

class BN;

class BNSS : public SS {
	friend class BN;
	SOBJCLASSDECL(BNSS)
public:
	BNSS(std::istream &is);
	BNSS(const BNSS &bnss);

	BNSS &operator=(const BNSS &bnss);
	virtual BNSS *Clone() const;

	virtual ~BNSS();

	virtual void Load(std::istream &is);
	virtual void Save(std::ostream &os) const;
        virtual void Scale(double w);
        virtual void AddSS(const SS* nss, double w=1.0);
private:
	BNSS();
	std::vector<SS *> rvss;
};


// BN is a class to represent a Bayesian Network
class BN : public RV {
	SOBJCLASSDECL(BN)
friend class ExactMarkovInf;
public:
	BN();
	BN(std::istream &is);
	BN(const BN &bn);
	BN(const Structure &s, const Context &vars);

	BN &operator=(const BN &bn);
	virtual BN *Clone() const;

	virtual ~BN();

	virtual void Load(std::istream &is);
	virtual void Save(std::ostream &os) const;

	virtual void Normalize();

	virtual double Prob(const Instantiation &x, bool log=false) const;

	virtual void Restrict(const Instantiation &x);

	virtual void Project(const Context &c, const Context &cc);

	virtual void MultBy(const RV *x);

	virtual BNSS *BlankSS() const;

	virtual void AddSS(const Instantiation &x, 
			SS *ss, double w=1.0) const;
	virtual void AddExpSS(const Instantiation &x, 
			SS *ss, double w=1.0) const;
	virtual void AddExpSS(const Instantiation &x, const RVSimple *rvs,
			SS *ss, double w=1.0) const;

	virtual void AddSS(const SS* toadd, const RV *rv,
			SS *ss, double w=1.0) const {}

	virtual void Sample(Instantiation &x, Random &rand=randomizer) const;

	virtual void Maximize(const SS *ss);

	virtual RVSimple *MakeSimple(const Instantiation &x, 
			bool normalize=true) const;

	// the BN now owns the pointer rv (call with rv->Clone()) if the
	// caller doesn't want to give up the variable
	void AddNode(RV *rv);

	// implements importance sampling and returns the weight of the
	// sample
	double ImportanceSample(Instantiation &x, 
			bool log=false, Random &rand=randomizer) const;

	virtual void Scramble(double alpha=1.0, double degree=1.0, 
			Random &rand=randomizer);

	virtual double LLH(const SS *ss) const;

	virtual double GetScore(double numTrans, const SS* ss) const;

	void FillParams(const BN &src);

	const RV *Node(int i) const {return nodes[i]; }
	const RV *NodeByVar(int i) const;
	int NumofNodes() const { return nodes.size(); }
        void GetStructure(Structure &s) const;

protected:
	void SetSampleOrder() const;
	std::vector<RV *> nodes;
	mutable std::vector<int> sampleorder; // as a cache
	// returns joint distribution restricted to x
	RV *Joint(const Instantiation &x) const;
	// generates the items below
	// call before use -- will do nothing if they already exist
	void PrepNode2IdMaps() const;
	// clears them (call if nodes changes)
	void ClearNode2IdMaps() const;

	// cache of mapping from index into "nodes" to the variable id represented
	mutable std::vector<int> node2var;
	// cache of mapping from variable id to the index into "nodes"
	mutable std::map<int,int> var2node;
	// cache of the indexes into "nodes" of the children for a particular
	// variable id (Note this one is slightly odd as it maps a variable id into
	// a list of node-indexes (not back into varids)
	mutable std::map<int,std::vector<int> > children;
};

#endif
