/* Continuous Time Bayesian Network Reasoning and Learning Engine
 * Copyright (C) 2009 The Regents of the University of California
 *
 * Authored by Yu Fan, William Lam, Joon Lee, Christian Shelton, and Jing Xu
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef CLIQUE_H_
#define CLIQUE_H_

#include "bn.h"
#include "markov.h"
#include "dynamics.h"
#include "context.h"
#include "ctbndyn.h"
#include "markovdyn.h"
#include "streamextra.h"
#include "markov.h"
#include <iostream>
#include <sstream>
#include <set>
#include <ctime>
#include <cstdlib>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include "random.h"


//clique tree will be use to improve process and decision process of CTBN
//this will build the clique tree base on the CTBNDyn, Markov, BN
//output will be vector of clique::cliques

class CliqueTree: public StreamObj{
public:

	CliqueTree();

	//build the clique tree for a markov (both CTBNDyn and BN factors)
	void BuildClique(const Markov &process, Random &rand = randomizer);
	//build the clique tree for a CTBNDyn
	void BuildClique(const CTBNDyn *dyn, Random &rand = randomizer);
	//build the clique tree for a BN
	void BuildClique(const BN *bn, Random &rand = randomizer);

	//print the clique tree
	void CliquePrint(std::ostream &os);


	//save the current state
	virtual void Save(std::ostream &os) const;


	//loading the value into the class so it will have all the data
	virtual void Load(std::istream &is);

	//to use for testing:
	//if true is passed in random number generator is ignored
	//and factors are processed in a deterministic order
	void IgnoreRand(bool ignore) {
		unit_test = ignore;
	}

	//change the number of random restarts
	void SetNumRestarts(int num) {
		numrandrestarts = num;
	}

	class Node {
	public:
		Context vars; // variables that are part of this factor
		std::vector<int> adj; // index of adjacent nodes in tree
	};

	//return the clique tree
	std::vector<Node> ReturnCliques() const;

	virtual ~CliqueTree();

private:
	class CliqueNode {
		friend class CliqueTree;
	public:
		CliqueNode();
		void AddVar(const int &);
		void print (std::ostream &);
		~CliqueNode();
	private:
		std::set <int> connected;
		std::vector <CliqueNode *> next;
		CliqueNode * prev;
	};

	void CliqueTrees(Random &rand);
	void CliqueTrees(int);
	void CliqueNodePrint(CliqueNode *, int, std::ostream &);
	static void Destroy(CliqueNode *);
	void ReturnClique(CliqueNode *, int, std::vector<Node> &) const;
	void InsertCliqueNode(CliqueNode *, int, std::vector<Node> &) const;
	Context Clique2Context(CliqueNode *) const;
	void CliqueSave(std::ostream &os, CliqueNode * print, int size) const;
	void Eliminate();
	bool EliminateDups(CliqueNode *);
	void EliminatePrev(CliqueNode *);
	static void RemoveElem(std::vector <int> &, int);
	void prints(std::set <int>, std::ostream &) const;
	double MaxSize() const;
	double MaxSize(CliqueNode *) const;
	void Dequeue(std::vector <int> remove, CliqueNode * newitem);

	void BuildIt(Random &);
	void AddFromBN(const BN *);
	void AddFromCTBNDyn(const CTBNDyn *);

	std::vector <CliqueNode> data;
	std::vector <CliqueNode> save_data;
	std::vector <CliqueNode *> clique_builded;
	CliqueNode * Head_compare;
	bool exist;
	int numrandrestarts;
	CliqueNode * Head;
	bool unit_test;
	int VarSize(int varid) const;
	void AddVarSize(int varid, int size);
	// maps varid -> size
	std::map<int,int> context_size;
};

#endif /* CLIQUE_H_ */
