/* Continuous Time Bayesian Network Reasoning and Learning Engine
 * Copyright (C) 2009 The Regents of the University of California
 *
 * Authored by Yu Fan, William Lam, Joon Lee, Christian Shelton, and Jing Xu
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef _H_context
#define _H_context

#include <vector>
#include <map>
#include <numeric>

class Instantiation;

class Context {
public:
	Context(); // assumes null context

	enum setop {UNION, INTERSECTION, DIFFERENCE};
	Context(const Context &c1, const Context &c2, setop op=UNION);
	// creates a context of the variables that are difference between
	//  the two instantiations (variables that are "unmentioned"
	//  in one or the other are not included)
	Context(const Instantiation &i1, const Instantiation &i2);
	Context(std::istream &is);

	virtual ~Context();

	int Index(const Instantiation &ind) const;
	Instantiation Index(int i) const;

	// assumes the varid is not already there...
	// if so, future behavior is undefined
	void AddVar(int varid, int cardinality) {
		size*=cardinality;
		card[varid]=cardinality;
	}

	std::vector<int> VarList() const;

	int Size() const {
		return size;
	}

	int NumVars() const { return card.size(); }
	int MinVar() const { return card.empty() ? -1 : card.begin()->first; }
	int MaxVar() const { return card.empty() ? -1 : card.end()->first; }
	bool HasId(int varid) const { return card.find(varid) != card.end(); }
	int Cardinality(int varid) const {
		std::map<int,int>::const_iterator l = card.find(varid);
		return l != card.end() ? l->second : 0;
	}

	void InconsistentIndexes(std::vector<int> &l, const Instantiation &v) const;
	void ConsistentIndexes(std::vector<int> &l, const Instantiation &v) const;

	virtual std::ostream &Save(std::ostream &os) const;
	virtual std::istream &Load(std::istream &is);

	// returns whether c is a subset (not proper) of *this
	bool IsSubset(const Context &c) const;
	bool IsOverlap(const Context &c) const;

protected:
	std::map<int,int> card; // maps id to max value
	int size;

	// These "complete" the calculation (and are reuseable in other
	//  similar calculations)
	void CConsistentIndexes(std::vector<int> &l, const std::vector<int> &maxind, const std::vector<int> &incind, int startindex) const;
	void CInconsistentIndexes(std::vector<int> &l, const std::vector<int> &lneg) const;
};

// forms union:
inline Context operator+(const Context &c1, const Context &c2) {
	return Context(c1,c2,Context::UNION);
}
// forms difference
inline Context operator-(const Context &c1, const Context &c2) {
	return Context(c1,c2,Context::DIFFERENCE);
}
// forms intersection
inline Context operator/(const Context &c1, const Context &c2) {
	return Context(c1,c2,Context::INTERSECTION);
}

class Instantiation : public Context {
friend class Context;
public:
	Instantiation();  // assumes null context
	Instantiation(const Context &c, int defaultval=-1);
	Instantiation(const Context &c, const Instantiation &v);
	virtual ~Instantiation();

	int Value(int varid) const {
		std::map<int,int>::const_iterator l = vals.find(varid);
		return l != vals.end() ? l->second : -1;
	}

	// returns false iff the increment has "wrapped"
	bool Inc(int varid);
	bool Inc();
	bool Inc(const Context &c);

	bool SetVal(int varid, int val) { // returns whether it was added
		std::map<int,int>::iterator l = vals.find(varid);
		if (l != vals.end()) {
			ChangeVar(varid,l->second,val);
			l->second = val;
			return true;
		} else return false;
	}

	void SetVal(const Instantiation &val, bool add=false) {
		for(std::map<int,int>::const_iterator l = val.vals.begin(),
				ll = val.card.begin();
				l!=val.vals.end();++l,++ll)
			if (!SetVal(l->first,l->second) && add)
				AddVar(l->first,ll->second,l->second);
	}

	void SetAllVal(int val);

	void AddVar(int varid, int cardinality, int val=-1);

	virtual std::ostream &Save(std::ostream &os) const;
	virtual std::istream &Load(std::istream &is);

	int Index() const { return numunknown ? -1 : index; }
	void SetIndex(int i);

	// used with ZeroOut from IMatrix (and ZeroOut above)
	void InconsistentIndexes(std::vector<int> &l) const;
	void ConsistentIndexes(std::vector<int> &l) const;

	Context KnownVars() const;
	Context MissingVars() const;
	int NumMissingVars() const { return numunknown; }
        
	void PrintVal(std::ostream &os) const; //debug use only

protected:
	void ChangeVar(int varid, int oldval, int newval) {
		if (oldval==-1) { numunknown--; oldval = 0; }
		if (newval==-1) { numunknown++; newval = 0; }
		index += indexinc[varid]*(newval-oldval);
	}
	void Recalc();

	std::map<int,int> vals;
	std::map<int,int> indexinc;
	int index, numunknown;
};

std::ostream& operator<< (std::ostream& os, const Instantiation& inst);
std::istream& operator>> (std::istream& is, Instantiation& inst);

std::ostream& operator<< (std::ostream& os, const Context& con);
std::istream& operator>> (std::istream& is, Context& con);

#endif
