/* Continuous Time Bayesian Network Reasoning and Learning Engine
 * Copyright (C) 2009 The Regents of the University of California
 *
 * Authored by Yu Fan, William Lam, Joon Lee, Christian Shelton, and Jing Xu
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef _DYNAMICS_H_
#define _DYNAMICS_H_

#include "context.h"
#include "trajectory.h"
#include "rvsimple.h"
#include "rv.h"

// Dynamics is a continuous-time semi-Markov process except for the
// starting distribution
// The domain is explicit
class Dynamics : public StreamObj {
public:
	Dynamics(const Context &var, const Context &cvar);
	Dynamics(std::istream &is);
	virtual ~Dynamics();

	virtual void Load(std::istream &is);
	virtual void Save(std::ostream &os) const;

	virtual Dynamics *Clone() const = 0;

	const Context &Domain() const { return v; }
	const Context &CondDomain() const {return cv; }

	virtual void Normalize() = 0;

	virtual void Restrict(const Instantiation &x) = 0;

	// the version here just sets up the contexts (v and cv)
	virtual void Mult(const Dynamics *x);

	// the implied domain for RVCondSimple is the Context
	// of this dynamic process
	// x gives the values for the conditional context
	// as well as any subsystem restrictions for the process's context
	virtual RVCondSimple *Cond(double t0, double t1,
			const Instantiation &x) const = 0;
	// same, but instantaneous intensity of transitioning
	// if transition==false, this doesn't correspond to a
	// transition, but rather just a change of support
	virtual RVCondSimple *Cond(double t, const Instantiation &from,
		const Instantiation &to, bool transition=true) const = 0;

	virtual SS *BlankSS() const = 0;

	virtual void AddSS(const Instantiation &x, double t0,
			double deltat, SS *ss, double w=1.0) const = 0;
	virtual void AddTransSS(const Instantiation &x1, 
				const Instantiation &x2,
				double t, SS *ss, double w=1.0) const = 0;

	virtual void AddExpSS(const RV *alpha, const RV *beta,
				double t0, double deltat, 
				SS *ss, double w=1.0) const = 0;
	virtual void AddExpTransSS(const RV *x1, const RV *x2,
			const Context &changevar,
			double t, SS *ss, double w=1.0) const = 0;

	// These are the same as above, but where the context is implied
	// and the conditioning variable takes on a known value (not
	// a distribution over them) -- either its index or a full 
	// instantiation is supplied...
	virtual void AddExpSS(const Instantiation &condi,
			const RVSimple *alpha, const RVSimple *beta,
			double t0, double deltat, SS *ss, 
			double w=1.0) const = 0;
	virtual void AddExpSS(int condi,
			const RVSimple *alpha, const RVSimple *beta,
			double t0, double deltat, SS *ss, 
			double w=1.0) const = 0;
	virtual void AddExpTransSS(const Instantiation &condi,
			const RVSimple *x1, const RVSimple *x2,
			const Context &changevar,
			double t, SS *ss, double w=1.0) const = 0;
	virtual void AddExpTransSS(int condi,
			const RVSimple *x1, const RVSimple *x2,
			const Context &changevar,
			double t, SS *ss, double w=1.0) const = 0;

	// This adds the sufficient statistics of a different process
	// to this process's sufficient statistics
	virtual void AddSS(const SS *toadd, const Dynamics *dyn,
			SS *ss, double w=1.0) const = 0;

	// assumes that the context of nexti equals Domain
	virtual void SampleNextEvent(const  Instantiation &i, double t,
			double &nextt, Instantiation &nexti,
			Random &rand=randomizer) const = 0;
	// completes the trajectory, starting at t and going until 
	// tr.EndTime()
	// the method implemented here repeatedly calls SampleNextEvent
	// override it if you can do something more efficient
	virtual void SampleTrajectory(Trajectory &tr, double t,
			Random &rand=randomizer) const;

	virtual void Maximize(const SS *ss) = 0;

        virtual void Scramble(double a=1.0, double b=1.0, 
				double alpha=1.0, double degree=1.0, 
				Random &rand=randomizer) = 0; //by Yu
        virtual double LLH(const SS *ss) const = 0;

	SS* SuffStats(const std::vector<Trajectory> &tr, 
			const std::vector<double> &w) const;
	virtual double GetScore(double numTrans, double amtTime, 
				const SS *ss) const = 0;
protected:
	Context v,cv;

};

#endif
