/* Continuous Time Bayesian Network Reasoning and Learning Engine
 * Copyright (C) 2009 The Regents of the University of California
 *
 * Authored by Yu Fan, William Lam, Joon Lee, Christian Shelton, and Jing Xu
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef _EXACTMARKOVINF_H_
#define _EXACTMARKOVINF_H_

#include "markov.h"
#include "ctbndyn.h"
#include "dyncomp.h"
#include "bn.h"
#include "rvcomp.h"
#include "inference.h"
#include <vector>

// This class performs exact inference by using the conditional distributions
// supplied by the process's (assumed to be a Markov process) dynamics's
// method "Cond."  For compact structures (like a CTBN) this will probably
// require generation of the full "flattened" dynamics which is exponentially
// (in the number of variables) large.  However, for some processes (like a
// fully factored process), it might be possible to return a conditional
// object that can represent this more compactly.
//
// The dynamics's "cond" methods should return RVCondSimple objects that
// are compatible with the RVSimple object's "MakeSimple" method (in that
// the RVCondSimple can accept the output of the MakeSimple method).

class ExactMarkovInf : public Inference {
	// (in fact, Inference isn't even a streamable object...
	// Loading and Saving are
	// not currently allowed as this class does not have
	// ownership over the Markov process and Trajectory
	// to which it has pointers/references
	//SOBJCLASSDECL(ExactMarkovInf)
public:
	ExactMarkovInf();
	virtual ~ExactMarkovInf();

	virtual ExactMarkovInf *Clone() const;

	virtual void SetProcess(const Process *p);
	virtual void SetTrajectory(const Trajectory *tr);

	virtual double Filter(const Instantiation &x, double t,
				bool log=false);
	virtual double Smooth(const Instantiation &x, double t,
				bool log=false);

	virtual double Prob(double t, bool log=false);

	virtual void AddExpSuffStats(SS *ss, double w=1.0);
	virtual void AddExpSuffStats(const Dynamics *dyn, SS *ss,
							double w=1.0);
	virtual void AddExpSuffStats(const RV *p0, SS *ss,
							double w=1.0);

	virtual void printdist(std::ostream &os) const {p->Save(os);};
	double CalcQuery(QueryCalculator &calc);

private:

	double FilterReal(const Instantiation &x, double t,
			bool log=false, bool normalize=true);
	const Markov *p;
	const Trajectory *tr;

	void ClearCache();
	// make up until t
	void MakeCache();
	void GetProp();
	void AlphaPass();
	void BetaPass();
	// the cache:
	std::vector<double> times;
	std::vector<RVSimple *> alpha;
	std::vector<RVSimple *> beta;
	std::vector<RVCondSimple *> prop;
	std::vector<int> condi;
	std::vector<int> transtype;
	std::vector<Context> changevars;

	// Cache the joint SS
	SS* jointP0SS;
	SS* jointDSS;
};

#endif

