/* Continuous Time Bayesian Network Reasoning and Learning Engine
 * Copyright (C) 2009 The Regents of the University of California
 *
 * Authored by Yu Fan, William Lam, Joon Lee, Christian Shelton, and Jing Xu
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef _GIBBSSAMPLER_H_
#define _GIBBSSAMPLER_H_

#include <map>
#include <vector>
#include "sampler.h"

class matrix;
class MultiSimple;
class RV;
class MarkovDyn;

// GibbsSampler is an implementation of the paper:
// Tal El-Hay, Nir Friedman, and Raz Kupferman (2008). "Gibbs Sampling in
// Factorized Continuous-Time Markov Processes."
// This class will take in an evidence trajectory, a markov process, 
// the number of samples needed as parameters.
// It will first burn in for several iterations (number of burnin_iter is
// specified in construction function), then generate a set of sampled 
// trajectories. It is an approximate inference method used for learning 
// and inference on CTBNs.

// Sample usage:
//
// Markov ctbn;
// Trajectory evidence;
// GibbsSampler sampler(&ctbn, &evidence);
// Trajectory sample_traj;
// sampler.SampleTrajectory(sample_traj);

class GibbsSampler : public Sampler
{
 public:
	GibbsSampler(const Process *process, const Trajectory *evidence, 
				 int burnin);
	virtual ~GibbsSampler() { if (init_traj) delete init_traj; }

	// inherited interface functions from base class
	virtual Sampler *Clone() const {
		return new GibbsSampler(p, evid, numBurninIter); 
	}
	virtual void SetProcess(const Process *process) { 
		p = process; Initialize(); 
	}
	virtual void SetTrajectory(const Trajectory *evidence) { 
		evid = evidence; 
	}
	virtual void SetMethod(const VarSample *m) { method = m; }
	virtual void SampleTrajectories(std::vector<Trajectory> &traj, 
					std::vector<double> &w,
					int numsamples, 
					Random &rand = randomizer);


 protected:
	// specific public interfaces for GibbsSampler
	// burn in, do numBurninIter steps of markov transitions
	void BurnIn(Random &rand = randomizer) const;

	// one step in the markov chain state transition; 
	virtual void Next(int num_iter = 1, Random &rand = randomizer) const;

	// get the current trajectory
	const Trajectory &Get() const { return tr; }
	// get the start point - the initial trajectory
	const Trajectory &GetInitTraj() const { return *init_traj; }
	// set the initial trajectory
	void SetInitTraj(const Trajectory &t);
	// delete the initial trajetcory
	void ClearInitTraj();



	int numBurninIter; // # of burn in iterations
	Context context;
	std::vector<int> own_var_list;
	const RV *p0;
	const MarkovDyn *md;
	// mutable so that we can access parents[v] 
	// etc. easily in const member fuctions.
	// each variable's parents and children
	mutable std::map<int, Context> parents, children;
	// each variable's markovdyn
	mutable std::map<int, const MarkovDyn *> mds; 
	// initial trajectory
	mutable Trajectory *init_traj; 
	// the current sample
	mutable Trajectory tr; 

	// inherited functions from base class, currently do nothing
	double SampleInitial(Instantiation &inst, Random &rand=randomizer){return 0.0;};
	double SampleDyn(Trajectory &traj, Random &rand=randomizer){return 0.0;};

	// precompute context, p0, dynamics, parents/children and markovdyn
	// for each variable;
	void Initialize();
	// returns true if c has variable in v's Markov blanket
	bool IsInBlanket(const Context &c, int v) const;
	// Find the {tau} splitting points (timing, full instantiation) for 
	// transitioned variables who are in v's Markov blanket
	void GetTau(const Trajectory &t, double begin_time, double end_time, 
				int v, std::vector<double> &tau, 
				std::vector<Instantiation> &inst) const;
	// Resample the entire trajectory of v given all the other variables'
	// full trajectory.
	virtual void SampleVariable(int v, Random &rand = randomizer) const;
	// Sample initial trajectory that agrees with evidence *evid.
	virtual void SampleInitialTrajectory(Random &rand = randomizer) const;
	
 private:
	// find evidence on v, and split into "free" intervals for v
	// there is no v's evidence inside each [t0, tT] interval
	void SplitFreeIntervals(int v, 
				std::vector<int> &x0, 
				std::vector<int> &xT, 
				std::vector<double> &t0, 
				std::vector<double> &tT, 
				const Instantiation &inst0, 
				Random &rand = randomizer) const;
	// sampler v's trajectory for interval [t0, tT];
	void SampleVariableInterval(int v, int x0, int xT, 
					double t0, double tT,
					Random &rand = randomizer) const;
	// equivalent to Qss;
	matrix GetRMatrix(int v, const Instantiation &inst) const;
	// equivalent to Qss';
	matrix GetTMatrix(int v, const Instantiation &oldinst, 
					  const Instantiation &newinst) const;
	// search the exact time points between t0 and t1 using binary search;
	double SearchPoint(double ksi, int x0, const matrix &R, 
				double log_p_future0, double log_p_past, 
				const MultiSimple &p_future, double log_z, 
				double t0, double t1, 
				double epsilon = 1e-6) const;
	// sample the initial bn conditioned on evindence "partial"
	Instantiation SampleConditionalRv(const Instantiation &partial, 
					int v,
					Random &rand = randomizer) const; 
};

#endif
