/* Continuous Time Bayesian Network Reasoning and Learning Engine
 * Copyright (C) 2009 The Regents of the University of California
 *
 * Authored by Yu Fan, William Lam, Joon Lee, Christian Shelton, and Jing Xu
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef _MARKOV_H_
#define _MARKOV_H_

#include "process.h"
#include "dynamics.h"
#include "rv.h"
#include "ss.h"
#include "structure.h"
#include "extramath.h"

class ExactMarkovInf;

class Markov : public Process {
	friend class ExactMarkovInf;
	friend class SamplingInf;
	SOBJCLASSDECL(Markov)
public:
	// this now "owns" these two pointers
	// so make sure they aren't on the stack!
	Markov(RV *startdist, Dynamics *dyn);
	Markov(std::istream &is);
	Markov(const Markov &m);
	Markov &operator=(const Markov &m);
	virtual ~Markov();

	virtual Markov *Clone() const;

	virtual void Mult(const Process *p);
	virtual SS *BlankSS() const;
	virtual void Maximize(const SS *ss);

	virtual void Load(std::istream &is);
	virtual void Save(std::ostream &os) const;

	virtual void Sample(Trajectory &tr, Random &rand=randomizer) const;

	const Dynamics *GetDynamics() const { return d; }
	const RV *GetStartDist() const {return p0;}

	double LLH(const Trajectory &tr, double w, 
			double &p0llh, double &dynllh) const;
	double LLH(const std::vector<Trajectory> &tr, 
			const std::vector<double> &w) const; 
	virtual double LLH(const SS *ss) const;
	SS* SuffStats(const std::vector<Trajectory> &tr, 
			const std::vector<double> &w) const;
	virtual void Scramble(double a=1.0, double b=1.0, 
				double alpha=1.0, double degree=1.0, 
				Random &rand=randomizer);
	
	virtual double GetScore(double numTrans, double amtTime, 
				const SS* ss) const;

protected:
	RV *p0;
	Dynamics *d;
};

class MarkovSS : public SS {
	friend class Markov;
	friend class ExactMarkovInf;
	friend class SamplingInf;
	SOBJCLASSDECL(MarkovSS)
public:
	MarkovSS(std::istream &is);
	MarkovSS(SS *rvss, SS *dynss);
	MarkovSS(const MarkovSS &mss);
	MarkovSS &operator=(const MarkovSS &mss);

	virtual ~MarkovSS();

	virtual MarkovSS *Clone() const;

	virtual void Load(std::istream &is);
	virtual void Save(std::ostream &os) const;
	virtual void Scale(double w);
	virtual void AddSS(const SS* nss, double w=1.0);
	virtual double NodeSS(int id, int val1, int val2, 
						  const Instantiation &cond) const;
private:
	SS *p0ss,*dss;
};

#endif

