/* Continuous Time Bayesian Network Reasoning and Learning Engine
 * Copyright (C) 2009 The Regents of the University of California
 *
 * Authored by Yu Fan, William Lam, Joon Lee, Christian Shelton, and Jing Xu
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef _MARKOVSIMPLE_H_
#define _MARKOVSIMPLE_H_

#include "dynsimple.h"

class MarkovSimple;

// The sufficient statistics for MarkovSimple:
class MarkovSimpleSS : public SS {
	SOBJCLASSDECL(MarkovSimpleSS)
  public:
	friend class MarkovSimple;
	friend class MarkovSimpleToggle;
	MarkovSimpleSS(int n);
	MarkovSimpleSS(std::istream &is);

	virtual ~MarkovSimpleSS();

	virtual MarkovSimpleSS *Clone() const;
	virtual void Load(std::istream &is);
	virtual void Save(std::ostream &os) const;
	virtual void Scale(double w);
	virtual void AddSS(const SS* nss, double w=1.0);
	virtual double Element(int s1, int s2) const {return c[s1][s2];}
  private:
	// diagonal elements hold the amount of time in each state
	// off-diagonal element (i,j) is the number of transitions from i to j
	matrix c;
};

// The next 4 classes are special versions of conditional random variables
// (mapping from a domain to the same domain).  They are used by MarkovSimple
// and are not designed to be learned from data or really interact with
// sufficient statistics.  Therefore all of the suffstats methods are
// blank.  For similar reasons they do not have serialization (load/save)
// currently.

// The next 2 classes represent the conditional distribution of the value of a
// Markov process at one time point given its value at a previous time point
// They store the Q matrix and the time interval instead of the stochastic
// matrix explicitly.
// When "Mult" is called, they generate the marginal distribution at the
// later point given the previous point (and the reverse for RMult) using
// numeric integration (see rk.h)

// This one is directly as described above.
template<typename QTYPE>
class CondTransQ : public RVCondSimple {
	//SOBJCLASSDECL(CondTransQ)
  public:
	CondTransQ(const matrix &QQ, double deltat);
	//CondTransQ(std::istream &is);
	virtual ~CondTransQ();
	virtual CondTransQ<QTYPE> *Clone() const;
	virtual double Prob(int ind, int cind, bool log=false) const;
	virtual RVSimple *Condition(int ind) const;
	virtual void Mult(RVSimple *&x) const;
	virtual void RMult(RVSimple *&x) const;
	virtual RVSimple *Convert(const RVSimple *rvs) const;

	virtual void Load(std::istream &is) {}
	virtual void Save(std::ostream &os) const {}
	virtual void Normalize() {}
	virtual void Restrict(const std::vector<int> &ind, 
				const std::vector<int> &cind) {}
	virtual void Reindex(int n, int nc, 
				const RVCondSimple::RemapT &ind) {}
	virtual void MultBy(const RVCondSimple *x) {}
	virtual SS *BlankSS() const { return NULL; };
	virtual void AddSS(int cind, int ind, SS *ss, double w=1.0) const {}
	virtual void AddExpSS(int cind, SS *ss, double w=1.0) const {}
	virtual void AddExpSS(int cind, const RVSimple *rvs, 
				SS *ss, double w=1.0) const {}
	virtual void AddSS(const SS *toadd, const RVCondSimple* rv,
			const std::vector<std::vector<int> > &mapping,
			int mycondi, int rvccondi, 
			SS *ss, double w=1.0) const {}
	virtual int Sample(int cx, Random &rand = randomizer) const;
	virtual void Maximize(const SS *ss) {}
	virtual void Scramble(double alpha=1.0, double degree=1.0, 
						  Random &rand=randomizer) {}
	virtual double LLH(const SS *ss) const {return 0.0;};
	virtual double GetScore(double numTrans, const SS *ss) const {return 0.0;};
  protected:
	QTYPE Q;
	double t;
};

// This one assumes that the system is constrained to a subset of the
// states (ind) and it only represents the submatrix of this state.
// It should be used with SparseMultiZSimple (as the RVSimple for Mult and
// RMult) to also capture this sparseness.  The methods will error if
// used with other versions of RVSimple
template<typename QTYPE>
class SparseCondTransQ : public RVCondSimple {
	//SOBJCLASSDECL(SparseCondTransQ)
  public:
	SparseCondTransQ(const matrix &QQ, double deltat, 
				const std::vector<int> &ind);
	virtual ~SparseCondTransQ();
	virtual SparseCondTransQ<QTYPE> *Clone() const;
	virtual double Prob(int ind, int cind, bool log=false) const;
	virtual RVSimple *Condition(int ind) const;
	virtual void Mult(RVSimple *&x) const;
	virtual void RMult(RVSimple *&x) const;
	virtual RVSimple *Convert(const RVSimple *rvs) const;

	virtual void Load(std::istream &is) {}
	virtual void Save(std::ostream &os) const {}
	virtual void Normalize() {}
	virtual void Restrict(const std::vector<int> &ind, 
				const std::vector<int> &cind) {}
	virtual void Reindex(int n, int nc, 
				const RVCondSimple::RemapT &ind) {}
	virtual void MultBy(const RVCondSimple *x) {}
	virtual SS *BlankSS() const { return NULL; };
	virtual void AddSS(int cind, int ind, SS *ss, double w=1.0) const {}
	virtual void AddExpSS(int cind, SS *ss, double w=1.0) const {}
	virtual void AddExpSS(int cind, const RVSimple *rvs, 
				SS *ss, double w=1.0) const {}
	virtual void AddSS(const SS *toadd, const RVCondSimple* rv,
			const std::vector<std::vector<int> > &mapping, 
			int mycondi, int rvccondi, 
			SS *ss, double w=1.0) const {}
	virtual int Sample(int cx, Random &rand = randomizer) const {return 0;}
	virtual void Maximize(const SS *ss) {}
	virtual void Scramble(double alpha=1.0, double degree=1.0, 
				Random &rand=randomizer) {}
	virtual double LLH(const SS *ss) const {return 0.0;}
	virtual double GetScore(double numTrans, const SS *ss) const {return 0.0;}
  protected:
	QTYPE Q;
	double t;
	std::vector<int> ix;
};

// The next two classes do the same thing (and again, for an "unconstrained"
// system and then for a "constrained" system).  However, they represent
// the conditional density of a transition at a time instance (therefore
// they map from "just before" the transition to "just after" the transition
// instead of across any non-zero interval of time).

// Here is the unconstrained (non-sparse) version.
template<typename QTYPE>
class CondTransQ2 : public RVCondSimple {
	//SOBJCLASSDECL(CondTransQ2)
  public:
	CondTransQ2(const matrix &QQ);
	virtual ~CondTransQ2();
	virtual CondTransQ2<QTYPE> *Clone() const;
	virtual double Prob(int ind, int cind, bool log=false) const;
	virtual RVSimple *Condition(int ind) const;
	virtual void Mult(RVSimple *&x) const;
	virtual void RMult(RVSimple *&x) const;
	virtual RVSimple *Convert(const RVSimple *rvs) const;

	virtual void Load(std::istream &is) {}
	virtual void Save(std::ostream &os) const {}
	virtual void Normalize() {}
	virtual void Restrict(const std::vector<int> &ind, 
				const std::vector<int> &cind) {}
	virtual void Reindex(int n, int nc, 
				const RVCondSimple::RemapT &ind) {}
	virtual void MultBy(const RVCondSimple *x) {}
	virtual SS *BlankSS() const { return NULL; };
	virtual void AddSS(int cind, int ind, SS *ss, double w=1.0) const {}
	virtual void AddExpSS(int cind, SS *ss, double w=1.0) const {}
	virtual void AddExpSS(int cind, const RVSimple *rvs, 
				SS *ss, double w=1.0) const {}
	virtual void AddSS(const SS *toadd, const RVCondSimple* rv,
			const std::vector<std::vector<int> > &mapping, 
			int mycondi, int rvccondi, 
			SS *ss, double w=1.0) const {}
	virtual int Sample(int cx, Random &rand = randomizer) const {return 0;}
	virtual void Maximize(const SS *ss) {}
	virtual void Scramble(double alpha=1.0, double degree=1.0, 
				Random &rand=randomizer) {}
	virtual double LLH(const SS *ss) const {return 0.0;};
	virtual double GetScore(double numTrans, const SS *ss) const {return 0.0;}
  protected:
	QTYPE Q;
};

// Here is the constrained version.  It is assumed that the state
// before transition is conditioned to be one of the cind set.
// The state after the transition is conditioned to be one of the ind set.
template<typename QTYPE>
class SparseCondTransQ2 : public RVCondSimple {
	//SOBJCLASSDECL(SparseCondTransQ2)
  public:
	SparseCondTransQ2(const matrix &QQ,
				const std::vector<int> &cind, 
				const std::vector<int> &ind);
	virtual ~SparseCondTransQ2();
	virtual SparseCondTransQ2<QTYPE> *Clone() const;
	virtual double Prob(int ind, int cind, bool log=false) const;
	virtual RVSimple *Condition(int ind) const;
	virtual void Mult(RVSimple *&x) const;
	virtual void RMult(RVSimple *&x) const;
	virtual RVSimple *Convert(const RVSimple *rvs) const;

	virtual void Load(std::istream &is) {}
	virtual void Save(std::ostream &os) const {}
	virtual void Normalize() {}
	virtual void Restrict(const std::vector<int> &ind, 
				const std::vector<int> &cind) {}
	virtual void Reindex(int n, int nc, 
				const RVCondSimple::RemapT &ind) {}
	virtual void MultBy(const RVCondSimple *x) {}
	virtual SS *BlankSS() const { return NULL; };
	virtual void AddSS(int cind, int ind, SS *ss, double w=1.0) const {}
	virtual void AddExpSS(int cind, SS *ss, double w=1.0) const {}
	virtual void AddExpSS(int cind, const RVSimple *rvs, 
				SS *ss, double w=1.0) const {}
	virtual void AddSS(const SS *toadd, const RVCondSimple* rv,
			const std::vector<std::vector<int> > &mapping, 
			int mycondi, int rvccondi, 
			SS *ss, double w=1.0) const {}
	virtual int Sample(int cx, Random &rand = randomizer) const {return 0;}
	virtual void Maximize(const SS *ss) {}
	virtual void Scramble(double alpha=1.0, double degree=1.0, 
				Random &rand=randomizer) {}
	virtual double LLH(const SS *ss) const {return 0.0;}
	virtual double GetScore(double numTrans, const SS *ss) const {return 0.0;}
  protected:
	QTYPE Q;
	std::vector<int> cix,ix;
};


class MarkovSimple : public DynSimple {
	SOBJCLASSDECL(MarkovSimple)
  public:
	MarkovSimple(int n=1);
	MarkovSimple(std::istream &is);
	virtual ~MarkovSimple();

	virtual void Load(std::istream &is);
	virtual void Save(std::ostream &os) const;

	virtual MarkovSimple *Clone() const;
	virtual MarkovSimple *MakeNew(int nstates) const;

	virtual void Normalize();
	virtual void Restrict(const std::vector<int> &ind);
	virtual void Expand(const std::vector<std::vector<int> > &ind, 
				int newn);

	virtual CondTransQ<const matrix &> *Cond(double t0, double t1) const;
	virtual SparseCondTransQ<matrix> *CondRestrict(double t0, double t1,
			const std::vector<int> &ind) const;

	virtual CondTransQ2<matrix> *Cond(double t) const;
	virtual SparseCondTransQ2<matrix> *CondRestrict(double t,
			const std::vector<int> &fromind,
			const std::vector<int> &toind, 
			bool transition=true) const;

	virtual MarkovSimpleSS *BlankSS() const;

	virtual void Mult(const DynSimple *x);

	virtual void AddSS(int x, double t0, double t1,
			SS *ss, double w=1.0) const;
	virtual void AddTransSS(int x1, int x2, double t,
			SS *ss, double w=1.0) const;
	virtual void AddExpSS(const RVSimple *alpha, const RVSimple *beta,
			double t0, double deltat, SS *ss, double w=1.0) const;
	virtual void AddExpTransSS(const RVSimple *x1, const RVSimple *x2,
			double t, SS *ss, double w=1.0) const;
	virtual void AddSS(const SS *toadd, const DynSimple *dyn,
			const std::vector<std::vector<int> > &mapping,
			SS *ss, double w=1.0) const;

	virtual void SampleNextEvent(int ind, 
					double t,
					double &newt, 
					int &newind, 
					Random &rand=randomizer) const;

	virtual void Maximize(const SS *ss);

	inline matrix &Intensity() { return Q; }
	inline const matrix &Intensity() const { return Q; }

	virtual void Scramble(double a=1.0, double b=1.0, 
				double alpha=1.0, double degree=1.0, 
				Random &rand=randomizer);
	virtual double LLH(const SS *ss) const;

	virtual double GetScore(int parentCard, double numTrans, 
			double amtTime, const SS* ss) const;
  private:
	matrix Q;
};

// added by Jing, handle toggle variables:
// as an extention to the MarkovSimple, but has parameter tying;
// e.g Q = [-2 1 1;
//          1 -2 1;
//          1 1 -2];
// a transition of a toggle variable only indicates a state changes;
// like a flip variable, not differentiate between changing from which to which;
// the probabilities of transit from one state to the other are all the same; 
// for a fixed size of the toggle variable, only 1 free paratmeter: the intensity of the state changes
class MarkovSimpleToggle : public MarkovSimple {
  public:
	MarkovSimpleToggle(int n=1);
	MarkovSimpleToggle(std::istream &is);
	virtual ~MarkovSimpleToggle();

	// calls MarkovSimple:MakeNew();
	virtual MarkovSimpleToggle *MakeNew(int nstates) const;
	// calls MarkovSimple:Clone();
	virtual MarkovSimpleToggle *Clone() const;
	// get the ML estimate from sufficient statistics ss; 
	// the state change intensity is estimated by total number of 
	// transitions between all pair of states divided by total number of 
	// time staying at all states
	virtual void Maximize(const SS *ss);
};


#endif
