/* Continuous Time Bayesian Network Reasoning and Learning Engine
 * Copyright (C) 2009 The Regents of the University of California
 *
 * Authored by Yu Fan, William Lam, Joon Lee, Christian Shelton, and Jing Xu
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef _STREAMEXTRA_H_
#define _STREAMEXTRA_H_

#include <math.h>
#include <string.h>
#include <algorithm>
#include "defines.h"

/* This header file does three separated (but related) things:
   1. It defines a way to save and load via streams (ie serialize)
      maps and vectors from STL.  Because of scoping issues it
	 has to be defined in the std namespace (which is unfortunate)

   2. It defines a function addnan which can be called on any
      input or output character stream.  It makes sure that stream
	 can read or write "NaN" and infinite values properly.

   3. It defines a SteamObj class.  This class can save and load
      any of its descendents to or from a stream without knowing the
	 type at compile time.  That is, you can load and save pointers
	 without having the object "sliced" to the type of the pointer.

	 Using it is slightly complex.  Basically, you need to define
	 a class that inherits (directly or indirectly) from StreamObj.
	 This class needs two things:
          i. a constructor from "std::istream &", and
	     ii. a virtual method Save that returns void and takes in
		      a "std::ostream &os"
      The former serves as a "load" method and the later as a save
	 method.

	 In the class declaration (on the first line) you need to
	 add the macro "SOBJCLASSDECL(xxx)" where xxx is the name of
	 the class.  Similarly, in the definition (.cc) file, you
	 need to add the macro "SOBJCLASSDEF(xxx)"

	 Below is a simple example.  The first constructor is not
	 required by the class.

	-------------in the .h file:

	class A : public StreamObj {
		SOBJCLASSDECL(A)
		
		public:
			A(int xx, int yy);
			A(std::istream &is);
			virtual void Save(std::ostream &os) const;

		private:
			int x,y;
	};

	--------------in the .cc file:

	SOBJCLASSDEF(A)

	A::A(int xx, int yy) {
		x = xx; y = yy;
	}

	A::A(std::istream &is) {
		is >> x >> y;
	}

	void A::Save(std::ostream &os) const {
		os << x << os.fill() << y;
	}

	---------------

	 To use the capabilities, simply save pointers to objects
	 by calling "SavePtr" and load them using StreamObj::LoadPtr.

	 For example:

	 A *aptr = new A(2,3);
	 aptr->SavePtr(cout);  // save object pointed to by aptr to cout
	 	// Note that if aptr pointed to a derived class of A,
		// provided it also had the form above, it would be saved
		// in its entirety (not just the "class A" portion).

	 StreamObj *objptr = StreamObj::LoadPtr(cin); // load obj from cin
	 	// Note that the resulting object will be of the 
		// "correct" type and not necessarily a StreamObj object.
*/

#include <map>
#include <vector>
#include <iostream>
#include <iterator>
#include <string>

// first declaration of some helper methods to push maps and vectors
// automatically through operator<< and operator>>
namespace std {
	template<typename T, typename S>
	ostream &operator<<(ostream &os, const pair<T,S> &p) {
		return os << p.first << os.fill() << p.second;
	}
	template<typename T, typename S>
	istream &operator>>(istream &is, pair<T,S> &p) {
		return is >> p.first >> p.second;
	}
	
	// basically the same as istream_iterator, but with a limit
	// on the number of items that can be read.
	template<typename T, typename CharT = char,
		typename Traits = char_traits<CharT>,
		typename Dist = ptrdiff_t>
	class istream_lim_iterator
	: public iterator<input_iterator_tag,T,Dist,const T*, const T&> {
		private:
			int togo;
			basic_istream<CharT,Traits> *s;
			T t;
		public:
			istream_lim_iterator() {
				togo = -1;
				s = NULL;
			}
			istream_lim_iterator(basic_istream<CharT,Traits> &is,
						 int limit)
					: s(&is), togo(limit) {
				read();
			}

			istream_lim_iterator& operator++() {
				read();
				return *this;
			}
			const istream_lim_iterator operator++(int) {
				istream_lim_iterator tmp = *this;
				(*this)++;
				return tmp;
			}
			const T operator*() const {
				return t;
			}
			const T* operator->() const {
				return &(operator*());
			}
			bool isequal(const istream_lim_iterator &ili) const {
				return (togo==-1 && ili.togo==-1) || 
						(togo!=-1 && ili.togo!=-1);
			}
		private:
			void read() {
				if (togo>=0) {
					if (togo>0 && s && s->good()) {
						*s >> t;
						togo--;
						if (!s || !s->good()) 
							togo = -1;
					} else togo=-1;
				}
			}
	};

	template<typename T, typename CharT, typename Traits, typename Dist>
	inline bool operator==(
			const istream_lim_iterator<T,CharT,Traits,Dist> &x,
			const istream_lim_iterator<T,CharT,Traits,Dist> &y) {
		return x.isequal(y);
	}

	template<typename T, typename CharT, typename Traits, typename Dist>
	inline bool operator!=(
			const istream_lim_iterator<T,CharT,Traits,Dist> &x,
			const istream_lim_iterator<T,CharT,Traits,Dist> &y) {
		return !x.isequal(y);
	}

	// And now we have the definitions of the shift operators:
	template<typename K, typename T, typename C, typename A>
	ostream &operator<<(ostream &os, const map<K,T,C,A> &m) {
		os << m.size() << os.fill();
		copy(m.begin(),m.end(),ostream_iterator<pair<K,T> >(os," "));
		return os;
	}

	template<typename K, typename T, typename C, typename A>
	istream &operator>>(istream &is, map<K,T,C,A> &m) {
		m.clear();
		int s;
		is >> s;
		m.insert(istream_lim_iterator<pair<K,T> >(is,s),
				istream_lim_iterator<pair<K,T> >());
		return is;
	}

	template<typename T, typename A>
	ostream &operator<<(ostream &os, const vector<T,A> &v) {
		os << v.size() << os.fill();
		copy(v.begin(),v.end(),ostream_iterator<T>(os," "));
		return os;
	}

	template<typename T, typename A>
	istream &operator>>(istream &is, vector<T,A> &v) {
		int s;
		is >> s;
		v.assign(istream_lim_iterator<T>(is,s),
				istream_lim_iterator<T>());
		return is;
	}

}

// Here is some code to define a new locale so that floating point
// numbers like -inf and Nan can be read and written.

// You can forget about all of it.  What you need to do is call
// this function (addnan) on any input or output stream before
// using it.  It will make sure that stream reads or writes "inf"
// and "NaN" values properly
void addnan(std::basic_ios<char> &str);


// This should work, but someone didn't add num_put_byname
// (there's time_put_byname!)
//template <class charT, class OutputIterator>
//class nannum_put : public std::num_put_byname<charT> {
template <class charT, charT *infstr, charT *nanstr, 
		class OutputIterator = std::ostreambuf_iterator<charT> >
class nannum_put : public std::num_put<charT,OutputIterator> {
	public:
		explicit nannum_put(const char *locname)
			//: std::num_put_byname<charT>(locname) {}
			: std::num_put<charT,OutputIterator>() {}
	protected:
		virtual OutputIterator do_put(OutputIterator itt, 
						std::ios_base &iob, 
						charT fill, 
						double d) const {
			if (!isfinite(d)) { 
				itt = writeit<double>(itt,d);
			} else {
				//itt = std::num_put_byname<charT>::
				//	do_put(itt,iob,fill,d);
				itt = std::num_put<charT,OutputIterator>::
						do_put(itt,iob,fill,d);
			}
			return itt;
		}
		virtual OutputIterator do_put(OutputIterator itt, 
				std::ios_base &iob, charT fill, 
				long double d) const {
			if (!isfinite(d)) { 
				itt = writeit<long double>(itt,d);
			} else {
				//itt = std::num_put_byname<charT>::
				//		do_put(itt,iob,fill,d);
				itt = std::num_put<charT,OutputIterator>::
					do_put(itt,iob,fill,d);
			}
			return itt;
		}
		virtual OutputIterator do_put(OutputIterator itt, 
						std::ios_base &iob, 
						charT fill, 
						float f) const {
			if (!isfinite(f)) { 
				itt = writeit<float>(itt,f);
			} else {
				//itt = std::num_put_byname<charT>::
				//		do_put(itt,iob,fill,f);
				itt = std::num_put<charT,OutputIterator>::
					do_put(itt,iob,fill,f);
			}
			return itt;
		}
	private:
		template<typename FP>
			static OutputIterator writeit(OutputIterator itt,
							FP f) {
				static const char *minus = "-";
				// commented-out code works for g++'s STL, but 
				// not for others.  Uncommented-out code is fine
				// more generally
				if (isinf(f)) {
					if (f<0) 
						//itt = std::__write(itt,"-",1);
						std::copy(minus,minus+1,itt);
					std::copy(infstr,
						infstr
						 +std::char_traits<charT>::length(infstr),
						itt);
					/*
					itt = std::__write(itt,
							infstr,
							std::
							char_traits<charT>::
							length(infstr));
							*/
				} else
					std::copy(nanstr,
						nanstr
						 +std::char_traits<charT>::length(nanstr),
						itt);
					/*
					itt = std::__write(itt,
							nanstr,
							std::
							char_traits<charT>::
							length(nanstr));
							*/
				return itt;
			}
};

// This should work, but someone didn't add num_get_byname
// (there's time_get_byname!)
//template <class charT, class InputIterator>
//class nannum_get : public std::num_put_byname<charT> {
template <class charT, charT *infstr, charT *nanstr, 
		class InputIterator = std::istreambuf_iterator<charT> >
class nannum_get : public std::num_get<charT,InputIterator> {
	public:
		typedef std::basic_string<charT> strtype;

		explicit nannum_get(const char *locname)
			//: std::num_get_byname<charT>(locname) {}
			: std::num_get<charT,InputIterator>() {}
	protected:
		virtual InputIterator do_get(InputIterator beg, 
						InputIterator end, 
						std::ios_base &iob, 
						std::ios_base::iostate &iost,
						 double &d) const {
			bool negate = false;
			if (*beg == '-') { negate=true; ++beg; }
			if (*beg == '+') { negate=false; ++beg; }
			if (*beg==*infstr) {
				if (checkstr(beg,end,infstr,iost))
					d = negate ? -INFINITY : INFINITY;
				return beg;
			} else if (*beg==*nanstr) {
				if (checkstr(beg,end,nanstr,iost))
					d = nan("");
				return beg;
			} else {
				InputIterator ret = 
					std::num_get<charT,InputIterator>::
						do_get(beg,end,iob,iost,d);
				if (negate) d = -d;
				return ret;
			}
		}
		virtual InputIterator do_get(InputIterator beg, 
						InputIterator end, 
						std::ios_base &iob, 
						std::ios_base::iostate &iost,
						 long double &d) const {
			bool negate = false;
			if (*beg == '-') { negate = true; ++beg; }
			if (*beg == '+') { negate = false; ++beg; }
			if (*beg==*infstr) {
				if (checkstr(beg,end,infstr,iost))
					d = negate ? -INFINITY : INFINITY;
				return beg;
			} else if (*beg==*nanstr) {
				if (checkstr(beg,end,nanstr,iost))
					d = nanl("");
				return beg;
			} else {
				InputIterator ret = 
					std::num_get<charT,InputIterator>::
						do_get(beg,end,iob,iost,d);
				if (negate) d = -d;
				return ret;
			}
		}
		virtual InputIterator do_get(InputIterator beg, 
						InputIterator end, 
						std::ios_base &iob, 
						std::ios_base::iostate &iost,
						float &f) const {
			bool negate = false;
			if (*beg == '-') { negate = true; ++beg; }
			if (*beg == '+') { negate = false; ++beg; }
			if (*beg==*infstr) {
				if (checkstr(beg,end,infstr,iost))
					f = negate ? -INFINITY : INFINITY;
				return beg; } else if (*beg==*nanstr) {
				if (checkstr(beg,end,nanstr,iost))
					f = nanf("");
				return beg;
			} else {
				InputIterator ret = 
					std::num_get<charT,InputIterator>::
						do_get(beg,end,iob,iost,f);
				if (negate) f = -f;
				return ret;
			}
		}
	private:
		static bool checkstr(InputIterator &beg, 
					InputIterator end, 
					charT *str, 
					std::ios_base::iostate &iost) {
			while(*str && beg!=end) {
				if (*beg==*str) { ++str; ++beg; }
				else {
					iost |= std::ios_base::failbit;
					return false;
				}
			}
			return *str==0;
		}
};
// End of stuff for addnan

// This is the StreamObj.  The macros below take care of all of the
// internals for inherited classes.  Do NOT overload any of these
// methods except Save!  Just define Save and a constructor (from 
// an input stream) for your new class and ignore the rest.
//
// The top of this file has instructions and an example
class StreamObj {
	public:
		StreamObj();
		virtual ~StreamObj();
		
		// This is the one that saves the info and should
		// be called on all ptrs
		void SavePtr(std::ostream &os) const;
		// This is standard (and the one you should overwrite)
		virtual void Save(std::ostream &os) const;

		// this one does all of the "RTTI-ish" stuff
		// this is the one you want to call to "load a ptr"
		static StreamObj *LoadPtr(std::istream &is);

		// this is the straight-forward one
		static StreamObj *Create(std::istream &is);

		typedef StreamObj* (*CreatorFn)(std::istream &);
		static const char *addID(const char *classname, CreatorFn fn);
		static const char *addID(const char *classname, 
					 	const char *classname2, 
						CreatorFn fn);
		static char *killSpaces(const char *classname);
		static const char *IDname;
	protected:
		virtual const char *myID() const;

	private:
		static std::map<std::string, CreatorFn> &loadtable();
};

#define SOBJCLASSDECL(classname) \
	protected: \
		virtual const char *myID() const; \
	public: \
		static const char *IDname; \
		static StreamObj *Create(std::istream &is);

#define SOBJCLASSDEF(classname) \
	const char *classname::IDname = StreamObj::addID(#classname,classname::Create); \
	StreamObj *classname::Create(std::istream &is) { \
		return new classname(is); \
	} \
	const char *classname::myID() const { return IDname; } \

//StreamObj::addID(#classname,tclass::IDname,classname<tclass>::Create); 
// This isn't that portable as typeid( ).name() isn't gauranteed to
// return anything useful and it certainly doesn't return the same thing
// across compilers.
// Another fix doesn't really seem possible just now.
// (asking class tclass for its name gets into a circular mess
//  where that name might not be set up yet)
#define SOBJCLASSDEFTEMPL1(classname,tclass) \
	template <class tclass> \
	const char *classname<tclass>::IDname = \
			StreamObj::addID(typeid(classname<tclass>).name(),classname<tclass>::Create); \
	template <class tclass> \
	StreamObj *classname<tclass>::Create(std::istream &is) { \
		return new classname<tclass>(is); \
	} \
	template <class tclass> \
	const char *classname<tclass>::myID() const { return IDname; } \

#endif

