/* Continuous Time Bayesian Network Reasoning and Learning Engine
 * Copyright (C) 2009 The Regents of the University of California
 *
 * Authored by Yu Fan, William Lam, Joon Lee, Christian Shelton, and Jing Xu
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <algorithm>
#include "gibbssampler.h"
#include "ctbndyn.h"
#include "markov.h"
#include "markovdyn.h"
#include "multisimple.h"
#include "bn.h"
#include "rk.h"
#include "utils.h"
#include "params.h"

using namespace std;

// "process" is a pointer to a (subclass of ) Process object.
// "evidence" is a pointer to Trajectory object that holds the context.
// Variables with value -1 means no evidence.
// Interval evidence X(t:t+dt)=x is represented as X(t)=x, X(t+dt)=-1.
// Point evidence X(t)=x is represented as X(t)=x, X(t+epsilon)=-1.
GibbsSampler::GibbsSampler(const Process *process, const Trajectory *evidence,
				 int burnin) {
	p = process;
	init_traj = NULL;
	evid = evidence;
	numBurninIter = burnin;
	Initialize();
}

void GibbsSampler::Initialize() {
	const Markov *ctbn = dynamic_cast<const Markov *>(p);
	const CTBNDyn *ctbndyn = dynamic_cast<const CTBNDyn *>
		(ctbn->GetDynamics());
	own_var_list = ctbndyn->Domain().VarList();
	context = ctbndyn->Domain() + ctbndyn->CondDomain();
	p0 = ctbn->GetStartDist();

	vector<int> var_list = context.VarList();
	for (size_t j=0; j!=var_list.size(); ++j) {
		// markovdyn for each variable
		const MarkovDyn *markov_dyn = dynamic_cast<const MarkovDyn *>
			(ctbndyn->NodeByVar(var_list[j]));
		Context domain = markov_dyn->Domain();
		vector<int> domain_var_list = domain.VarList();
		// domain_var_list should have exactly one variable
		mds[domain_var_list[0]] = markov_dyn;

		// parents and children set for each variable
		Context parent_context = Context();
		Context children_context = Context();
		for (int i=0; i<ctbndyn->NumofNodes(); ++i) {
			Context c = ctbndyn->Node(i)->Domain();
			Context cc = ctbndyn->Node(i)->CondDomain();
			if (c.HasId(var_list[j]))
				parent_context = parent_context + cc;
			if (cc.HasId(var_list[j]))
				children_context = children_context + c;
		}
		parents[var_list[j]] = parent_context;
		children[var_list[j]] = children_context;
	}
}

void GibbsSampler::SampleTrajectories(vector<Trajectory> &traj,
					vector<double> &w,
					int numsamples, Random &rand) {
	double t1 = getcputime();
	BurnIn(rand);
	double t2 = getcputime();
	times.clear();
	times.push_back(t2-t1);
	t1 = getcputime();
	for (int i=0; i<numsamples; ++i) {
		traj.push_back(Get());
		w.push_back(0.0);  // log weight
		Next();
	}
	t2 = getcputime();
	times.push_back(t2-t1);
}

void GibbsSampler::BurnIn(Random &rand) const {
	if (!init_traj) {
		SampleInitialTrajectory(rand);
	} else {
		tr = *init_traj;
	}
	Next(numBurninIter, rand);
}

void GibbsSampler::Next(int num_iter, Random &rand) const {
	for (int i=0; i<num_iter; ++i) {
		for (size_t var=0; var!=own_var_list.size(); ++var)
			SampleVariable(own_var_list[var], rand);
	}
}

void GibbsSampler::SetInitTraj(const Trajectory &t) {
	if (init_traj)
		delete init_traj;
	init_traj = new Trajectory(t);
	tr = *init_traj;  // also reset the current sample to init_traj
}

void GibbsSampler::ClearInitTraj() {
	if (init_traj) {
		delete init_traj;
		init_traj = NULL;
	}
}

void GibbsSampler::SampleInitialTrajectory(Random &rand) const {
	tr = Trajectory();
	tr.SetBeginTime(evid->TimeBegin());
	tr.SetEndTime(evid->TimeEnd());

	// Sample a start distribution from BN. The start distribution should
	// be consistent with the initial values in evidence.
	// NOTE: start with the next event after tr.Timebegin(), 
	// since the first one is already handled in BN's sample; 
	Trajectory::Index i = evid->Begin(context);
	Instantiation inst(i.Values());
	Instantiation inst0 = inst;
	p0->Sample(inst0, rand);
	tr.AddTransition(inst0, tr.TimeBegin());
	double dt = i.DeltaT();

	// loop over evidence
	for (++i; !i.Done(); ++i) {
		Instantiation newinst(i.Values());
		for (int v=0; v<context.NumVars(); ++v) {
			if (newinst.Value(v) == -1)
				newinst.SetVal(v, inst.Value(v));
		}

		Context diff(newinst, inst);
		vector<int> diff_vars = diff.VarList();
		for (size_t j=0; j!=diff_vars.size(); ++j) {
			double event_time = i.Time() - randomizer.RandReal() * dt;
			tr.AddTransition(diff_vars[j], event_time, 
						newinst.Value(diff_vars[j]));
		}
		inst = newinst;
		dt = i.DeltaT();
	}
}

void GibbsSampler::SampleVariable(int v, Random &rand) const {
	// Remove all v's transitions from tr.
	Trajectory newtr;
	newtr.SetBeginTime(tr.TimeBegin());
	newtr.SetEndTime(tr.TimeEnd());
	// deal with the first instantiation
	Trajectory::Index i = tr.Begin(context);
	Instantiation inst(i.Values());
	int oldv = inst.Value(v);
	inst.SetVal(v, -1);
	newtr.AddTransition(inst, i.Time());

	// looping over the whole trajectory
	for (++i; !i.Done(); ++i) {
		inst = i.Values();
		int newv = inst.Value(v);
		if (newv != oldv) {
			oldv = newv;
			continue;
		}
		inst.SetVal(v, -1);
		newtr.AddTransition(inst, i.Time());
	}
	tr = newtr;

	vector<int> x0, xT;
	vector<double> t0, tT;
	SplitFreeIntervals(v, x0, xT, t0, tT, 
				tr.Values(context, tr.TimeBegin()), rand);
	if (x0.empty())  // no free interval for v
		return;

	for (size_t i=0; i!=x0.size(); ++i)
		SampleVariableInterval(v, x0[i], xT[i], t0[i], tT[i], rand);
}

void GibbsSampler::SplitFreeIntervals(int v, 
					vector<int> &x0, 
					vector<int> &xT,
				 	vector<double> &t0, 
					vector<double> &tT, 
					const Instantiation &inst0, 
					Random &rand) const {
	// NOTE: if no evidence for v specifies the value of v at 
	// tr.TimeBegin() and tr.TimeEnd(), then we must sample 
	// x0 and xT for them.
	Trajectory::Index i = evid->Begin(context);
	Instantiation inst(i.Values());
	int oldval = inst.Value(v);
	if (oldval == -1) {
		// open free interval
		t0.push_back(i.Time());
		Instantiation init_vals = SampleConditionalRv(inst0, v, rand);
		x0.push_back(init_vals.Value(v));
		tr.AddTransition(v, tr.TimeBegin(), init_vals.Value(v));
	} else {
		tr.AddTransition(v, tr.TimeBegin(), oldval);
	}

	for (++i; !i.Done(); ++i) {
		inst = i.Values();
		int newval = inst.Value(v);

		if (newval!=oldval && newval!=-1)
			tr.AddTransition(v, i.Time(), newval);

		if (oldval==-1 && newval!=-1) {
			// close free interval
			tT.push_back(i.Time());
			xT.push_back(newval);
		} else if (oldval!=-1 && newval==-1) {
			// open free interval
			t0.push_back(i.Time());
			x0.push_back(oldval);
		}
		oldval = newval;
	}
	if (oldval == -1) {
		// close free interval
		tT.push_back(evid->TimeEnd());
		xT.push_back(-1);
	}
}

bool GibbsSampler::IsInBlanket(const Context &c, int v) const {
	if (parents[v].IsSubset(c) || children[v].IsSubset(c))
		return true;
	vector<int> children_list = children[v].VarList();
	for (size_t j=0; j!=children_list.size(); ++j) {
		if (parents[children_list[j]].IsSubset(c))
			return true;
	}
	return false;
}

void GibbsSampler::GetTau(const Trajectory &t, 
				double begin_time, 
				double end_time, 
				int v, 
				vector<double> &tau, 
				vector<Instantiation> &inst) const {
	Trajectory::Index index = t.Begin(context, begin_time, end_time);
	Instantiation oldinst(index.Values());
	tau.push_back(begin_time);
	inst.push_back(oldinst);
	vector<int> vids = context.VarList();

	for (++index; !index.Done(); ++index) {
		Instantiation newinst(index.Values());
		Context diff(oldinst, newinst);
		if (IsInBlanket(diff, v)) {
			inst.push_back(newinst);
			tau.push_back(index.Time());
		}
		oldinst = newinst;
	}

	tau.push_back(end_time);
	inst.push_back(oldinst);
}

// The implementation of the paper goes here.
void GibbsSampler::SampleVariableInterval(int v, int x0, int xT, 
						double t0, double tT, 
						Random &rand) const {
	static double kEpsilon = ParamDouble("GibbsEpsilon",1e-5);

	// interval too small
	if (fabs(tT-t0) < 2.0*kEpsilon) {
		if (x0==xT || xT==-1)  // already consistent
			return;
		tr.AddTransition(v, (t0+tT)/2.0, xT);
		return;
	}
	
	vector<double> tau;
	vector<Instantiation> inst;
	GetTau(tr, t0, tT, v, tau, inst);

	// Sample the time to stay at state x0.
	// Compute P(future) vectors for {tau} from K back to 0.

	vector<MultiSimple> p_future;  // normalized distribution vectors
	vector<double> log_z;  // log of the normalization factors of p_future
	double logfactor;  // required by GetDist but not used
	int tau_size = static_cast<int>(tau.size());
	int vdim = context.Cardinality(v);
	for (int i=0; i!=tau_size; ++i) {
		p_future.push_back(MultiSimple(vdim));
		log_z.push_back(0.0);
	}
	vectr dist(vdim, 0.0);
	if (xT == -1)
		dist = 1.0;
	else
		dist[xT] = 1.0;
	p_future[tau_size-1].SetDist(dist);
	for (int i=tau_size - 2; i>=0; --i) {
		// compute p_future[i] from p_future[i+1]
		matrix R = GetRMatrix(v, inst[i]);
		p_future[i+1].GetDist(dist, logfactor);
		logfactor = expmtv(dist, R, tau[i+1] - tau[i]);
		if (i > 0) {
			matrix T = GetTMatrix(v, inst[i-1], inst[i]);
			dist = T * dist;
		}
		p_future[i].SetDist(dist);
		logfactor += log(p_future[i].Normalize());
		log_z[i] = logfactor + log_z[i+1];
	}
	double log_p_future0 = p_future[0].Prob(x0, true) + log_z[0];
	double ksi = rand.RandReal();
	double log_p_past = 0.0;
	int which_interval;
	for (which_interval=0; which_interval<tau_size-1; ++which_interval) {
		double new_log_p_past = log_p_past;
		matrix R = GetRMatrix(v, inst[which_interval]);
		new_log_p_past += (tau[which_interval+1] - 
					tau[which_interval]) * R[x0][x0];
		double log_p_future = 
				p_future[which_interval+1].Prob(x0, true) + 
				log_z[which_interval+1];
		double Ft = 1.0 - 
			exp(new_log_p_past + log_p_future - log_p_future0);
		if (ksi < Ft)
			break;
		if (which_interval < tau_size-2) {
			matrix T = GetTMatrix(v, inst[which_interval], 
						inst[which_interval+1]);
			new_log_p_past += log(T[x0][x0]);
		}
		log_p_past = new_log_p_past;
	}
	// This happens only when xT == x0, in which case F(tT) < 1.
	if (which_interval >= tau_size - 1)
		return;
	matrix R = GetRMatrix(v, inst[which_interval]);
	double t = SearchPoint(ksi, x0, R, log_p_future0, log_p_past, 
				p_future[which_interval+1], 
				log_z[which_interval+1],
				tau[which_interval], 
				tau[which_interval+1], 
				kEpsilon);

	// stay till the end of trajectory
	if (fabs(tT-t) < 2.0*kEpsilon) {
		if (x0==xT || xT==-1)  // already consistent
			return;
		tr.AddTransition(v, (t+tT)/2.0, xT);
		return;
	}

	// Sample the next state to transition to.
	p_future[which_interval+1].GetDist(dist, logfactor);
	expmtv(dist, R, tau[which_interval+1] - t);
	for (int i=0; i<vdim; ++i)
		dist[i] *= R[x0][i];
	dist[x0] = 0.0;
	MultiSimple pft(vdim);
	pft.SetDist(dist);
	pft.Normalize();
	int next_state = pft.Sample(rand);

	// add the event to tr
	tr.AddTransition(v, t, next_state);
	SampleVariableInterval(v, next_state, xT, t, tT, rand);
}

matrix GibbsSampler::GetRMatrix(int v, const Instantiation &inst) const {
	const MarkovDyn *md = mds[v];
	matrix R = (*md)(inst)->Intensity();

	vector<int> children_id = children[v].VarList();
	for (size_t i=0; i!=children_id.size(); ++i) {
		int val = inst.Value(children_id[i]);
		Instantiation pinst = inst;
		const MarkovDyn *cmd = mds[children_id[i]];
		for (int j=0; j<R.getm(); ++j) {
			pinst.SetVal(v, j);
			R[j][j] += (*cmd)(pinst)->Intensity()[val][val];
		}
	}
	return R;
}

matrix GibbsSampler::GetTMatrix(int v, const Instantiation &oldinst, 
					const Instantiation &newinst) const {
	int vdim = context.Cardinality(v);
	matrix T(vdim, vdim, 0.0);
	Context diff(oldinst, newinst);

	// NOTE: assume there is exactly one variable changed
	int changeid = diff.VarList()[0];

	// if the changed variable is not a child of v; 
	// simply return an identity matrix;
	if (!children[v].HasId(changeid)) {
		for (int i = 0; i < vdim; ++i)
			T[i][i] = 1.0;
		return T;
	}
	// o.w. return a diagonal matrix whose T[i][i] = q(a->a')|i
	int a = oldinst.Value(changeid);
	int b = newinst.Value(changeid);
	const MarkovDyn *md = mds[changeid];
	Instantiation inst = oldinst;

	for (int i=0; i<vdim; ++i) {
		inst.SetVal(v, i);
		T[i][i] = (*md)(inst)->Intensity()[a][b]; //get the CIM
	}
	return T;
}

double GibbsSampler::SearchPoint(double ksi, 
					int x0, 
					const matrix &R, 
					double log_p_future0, 
					double log_p_past, 
					const MultiSimple &p_future, 
					double log_z, 
					double t0, 
					double t1, 
					double epsilon) const {
	// This loop is guaranteed to be terminating since each iteration we will
	// reduce (t1 - t0) by a half.
	for (;;) {
		double t = (t0+t1)/2.0;
		double dt = (t1-t0)/2.0;
		if (dt < epsilon)
			return t;
		// Compute F(t).
		vectr dist;
		double logfactor;
		p_future.GetDist(dist, logfactor);
		logfactor = expmtv(dist, R, dt);
		logfactor += log(dist.normalize());
		logfactor += log_z;

		double new_log_p_past = log_p_past;
		new_log_p_past += dt*R[x0][x0];
		MultiSimple new_p_future = p_future;
		new_p_future.SetDist(dist);
		double Ft = 1.0 - exp(new_log_p_past + log(dist[x0]) 
							  + logfactor - log_p_future0);

		if (fabs(Ft-ksi) < epsilon)
			return t;
		if (Ft < ksi) {  // search (t, t1)
			t0 = t;
		} else {
			t1 = t;  // search (t0, t)
		}
	}
}

Instantiation GibbsSampler::SampleConditionalRv(const Instantiation &partial,
						int v, Random &rand) const {
	double log_prob_e = p0->Prob(partial, true);
	int card_v = context.Cardinality(v);
	vector<double> posterior_prob(card_v);
	double max_log_prob = -1e20;
	for (int val=0; val<card_v; ++val) {
		Instantiation inst(partial);
		inst.SetVal(v, val);
		posterior_prob[val] = p0->Prob(inst, true);
		if (posterior_prob[val] > max_log_prob)
			max_log_prob = posterior_prob[val];
	}
	double summation = 0.0;
	for (int val=0; val<card_v; ++val) {
		posterior_prob[val] = exp(posterior_prob[val] - max_log_prob);
		summation += posterior_prob[val];
	}
	int sample_val = rand.SampleMultinomial(posterior_prob, summation);
	Instantiation inst = partial;
	inst.SetVal(v, sample_val);
	return inst;
	
}

