/* Continuous Time Bayesian Network Reasoning and Learning Engine
 * Copyright (C) 2009 The Regents of the University of California
 *
 * Authored by Yu Fan, William Lam, Joon Lee, Christian Shelton, and Jing Xu
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include<stdio.h>
#include<stdlib.h>
#include<math.h>
#include <vector>
#include "matrix.h"
#include "rk.h"
#include "params.h"

// This version has a calculation cache
// comment out the USECACHE definition to remove the caching
//#define USECACHE
#define MAXCACHE 1000

using namespace std;

#ifdef USECACHE
static int reuse = 0;
static int tries = 0;
#endif

// computes a * exp(Qt), via integration with desired error of eps
// (h, if non-negative, gives a starting estimate of step size)
// return value replaces a
// This is a version based on adaptive step sizes (and therefore
// a fifth-order Runge-Kutta method with an embedded 4th order method
// -- see Numerical Recipes in C)
double vexpmt(vectr &a, const matrix &Q, double t, double eps, double h) {
	double ret = 0.0;
	if (t<=0.0) return ret;

	static double defaulteps = ParamDouble("RKEPS",RKEPS);
	if (eps<0.0) eps = defaulteps;

#ifdef USECACHE
	static vector<vectr> cachea(MAXCACHE);
	static vector<matrix> cacheQ(MAXCACHE);
	static vector<double> cachet(MAXCACHE);
	static vector<vectr> cacheret(MAXCACHE);
	static vector<double> cachellh(MAXCACHE);
	static int cachepos = 0;
	tries++;
	//if (tries%1000==0) cout << "cache: " << reuse << '/' << tries << endl;
	for(int i=0;i<cachea.size();i++)
		if (cachea[i]==a && cacheQ[i]==Q && cachet[i]==t) {
			reuse++;
			a = cacheret[i];
			return cachellh[i];
		}
	cachea[cachepos] = a;
	cacheQ[cachepos] = Q;
	cachet[cachepos] = t;
#endif

	int p = Q.getm();

	if (h<=0) {
		double dd = Q.diag().absmax();
		if (dd>0) h = 0.1/Q.diag().absmax();
		else h = 0.1;
	}

	vectr ka1(p),ka2(p),ka3(p),ka4(p),ka5(p),ka6(p);

	//static double a2=0.2, a3=0.3, a4=0.6, a5=1.0, a6=0.875,
	static double
		b21=0.2,
		b31=3.0/40.0, b32=9.0/40.0,
		b41=0.3, b42=-0.9, b43=1.2,
		b51=-11.0/54.0, b52=2.5, b53=-70.0/27.0, b54=35.0/27.0,
		b61=1631.0/55296.0, b62=175.0/512.0, b63=575.0/13824.0,
		b64=44275.0/110592.0, b65=253.0/4096.0,
		c1=37.0/378.0, c3=250.0/621.0, c4=125.0/594.0, c6=512.0/1771.0,
		dc5=-277.00/14336.0;
	static double dc1=c1-2825.0/27648.0, dc3=c3-18575.0/48384.0,
		dc4=c4-13525.0/55296.0, dc6=c6-0.25;
        int iter=0;
	double currt = 0;
	while(currt<t) {
                if (currt+h>=t) h = t-currt; // don't overshoot!

		double err;
		for(;;) { // find stepsize
			// take 5th-order step:
			//iter++;
                    
			vectr temp;

			Q.leftmult(a,ka1);
			ka1 *= h;
			//ka1 = (a*Q)*h;

			temp = a;
			temp.add(ka1,b21);
			//temp = ka1*b21+a;
			Q.leftmult(temp,ka2);
			ka2 *= h;
			//ka2 = (temp*Q)*h;

			temp = a;
			temp.add(ka1,b31);
			temp.add(ka2,b32);
			//temp = ka1*b31+ka2*b32+a;
			Q.leftmult(temp,ka3);
			ka3 *= h;
			//ka3 = (temp*Q)*h;

			temp = a;
			temp.add(ka1,b41);
			temp.add(ka2,b42);
			temp.add(ka3,b43);
			//temp = ka1*b41+ka2*b42+ka3*b43+a;
			Q.leftmult(temp,ka4);
			ka4 *= h;
			//ka4 = (temp*Q)*h;

			temp = a;
			temp.add(ka1,b51);
			temp.add(ka2,b52);
			temp.add(ka3,b53);
			temp.add(ka4,b54);
			//temp = ka1*b51+ka2*b52+ka3*b53+ka4*b54+a;
			Q.leftmult(temp,ka5);
			ka5 *= h;
			//ka5 = (temp*Q)*h;

			temp = a;
               temp.add(ka1,b61);
               temp.add(ka2,b62);
               temp.add(ka3,b63);
               temp.add(ka4,b64);
               temp.add(ka5,b65);
               //temp = ka1*b61+ka2*b62+ka3*b63+ka4*b64+ka5*b65+a;
               Q.leftmult(temp,ka6);
               ka6 *= h;
               //ka6 = (temp*Q)*h;

			vectr erra = ka1;
               erra *= dc1;
               erra.add(ka3,dc3);
               erra.add(ka4,dc4);
               erra.add(ka5,dc5);
               erra.add(ka6,dc6);
			//vectr erra = dc1*ka1+dc3*ka3+dc4*ka4+dc5*ka5+dc6*ka6;

			// check error:
			/*err = erra.absmax();
			err /= eps;
			if (err <= 1.0) break;*/
                        double amin = a.absmin();
                        if(amin<1.0) amin = 1.0;
                        err = erra.absmax()/amin;
                        err /= eps;
                        if(err<=1.0) break;

			// adjust step:
			double htemp = 0.9*h*pow(err,-0.25);
			if (htemp>h/10) h = htemp;
			else h /= 10;
			if (currt+h == currt) {
                            /*cerr << "step size underflow in rk2" << endl;
				cerr << "in alpha pass:" << endl;
				cerr << Q << endl;
				cerr << a << endl;
				cerr << t << ' ' << currt << endl;
				exit(1);*/
                            return ret;
			}
		}
		currt += h;
		if (err > 1.89e-4) h = 0.9*h*pow(err,-0.2);
		else h *= 5;

		a.add(ka1,c1);
		a.add(ka3,c3);
		a.add(ka4,c4);
		a.add(ka6,c6);
		//a += c1*ka1+c3*ka3+c4*ka4+c6*ka6;
		double ss = a.sum();
		if (ss>0.0) {
			ret += log(ss);
			a /= ss;
		}
		//ret += log(a.normalize());
	}
#ifdef USECACHE
	cacheret[cachepos] = a;
	cachellh[cachepos] = ret;
	cachepos++;
	//if (cachepos>MAXCACHE) { cerr << "cache overflow" << endl; cachepos=0;}
	if (cachepos>=MAXCACHE) cachepos=0;
#endif
	return ret;
}


// computes exp(Qt) * b, via integration with desired error of eps
// (h, if non-negative, gives a starting estimate of step size)
// return value replaces b
// This is a version based on adaptive step sizes (and therefore
// a fifth-order Runge-Kutta method with an embedded 4th order-method
// -- see Numerical Recipes in C)
double expmtv(vectr &b, const matrix &Q, double t, double eps, double h) {
    double ret = 0.0;
	if (t<=0.0) return ret;

	static double defaulteps = ParamDouble("RKEPS",RKEPS);
	if (eps<0.0) eps = defaulteps;

#ifdef USECACHE
	static vector<vectr> cacheb(MAXCACHE);
	static vector<matrix> cacheQ(MAXCACHE);
	static vector<double> cachet(MAXCACHE);
	static vector<vectr> cacheret(MAXCACHE);
	static vector<double> cachellh(MAXCACHE);
	static int cachepos = 0;
	tries++;
	//if (tries%1000==0) cout << "cache: " << reuse << '/' << tries << endl;
	for(int i=0;i<cacheb.size();i++)
		if (cacheb[i]==b && cacheQ[i]==Q && cachet[i]==t) {
			reuse++;
			b = cacheret[i];
			return cachellh[i];
		}
	cacheb[cachepos] = b;
	cacheQ[cachepos] = Q;
	cachet[cachepos] = t;
#endif

	int p = Q.getm();

	if (h<=0) {
		double dd = Q.diag().absmax();
		if (dd>0) h = 0.1/Q.diag().absmax();
		else h = 0.1;
	}

	vectr kb1(p),kb2(p),kb3(p),kb4(p),kb5(p),kb6(p);

	//static double a2=0.2, a3=0.3, a4=0.6, a5=1.0, a6=0.875,
	static double 
		b21=0.2,
		b31=3.0/40.0, b32=9.0/40.0,
		b41=0.3, b42=-0.9, b43=1.2,
		b51=-11.0/54.0, b52=2.5, b53=-70.0/27.0, b54=35.0/27.0,
		b61=1631.0/55296.0, b62=175.0/512.0, b63=575.0/13824.0,
		b64=44275.0/110592.0, b65=253.0/4096.0,
		c1=37.0/378.0, c3=250.0/621.0, c4=125.0/594.0, c6=512.0/1771.0,
		dc5=-277.00/14336.0;
	static double dc1=c1-2825.0/27648.0, dc3=c3-18575.0/48384.0,
		dc4=c4-13525.0/55296.0, dc6=c6-0.25;

	double currt = 0;
	while(currt<t) {
		if (currt+h>=t) h = t-currt; // don't overshoot!

		double err;
		for(;;) { // find stepsize
			// take 5th-order step:
			vectr temp;

			Q.rightmult(b,kb1);
			kb1 *= h;
			//kb1 = (Q*b)*h;

			temp = b;
			temp.add(kb1,b21);
			//temp = kb1*b21+b;
			Q.rightmult(temp,kb2);
			kb2 *= h;
			//kb2 = (Q*temp)*h;

			temp = b;
			temp.add(kb1,b31);
			temp.add(kb2,b32);
			//temp = kb1*b31+kb2*b32+b;
			Q.rightmult(temp,kb3);
			kb3 *= h;
			//kb3 = (Q*temp)*h;

			temp = b;
			temp.add(kb1,b41);
			temp.add(kb2,b42);
			temp.add(kb3,b43);
			//temp = kb1*b41+kb2*b42+kb3*b43+b;
			Q.rightmult(temp,kb4);
			kb4 *= h;
			//kb4 = (Q*temp)*h;

			temp = b;
			temp.add(kb1,b51);
			temp.add(kb2,b52);
			temp.add(kb3,b53);
			temp.add(kb4,b54);
			//temp = kb1*b51+kb2*b52+kb3*b53+kb4*b54+b;
			Q.rightmult(temp,kb5);
			kb5 *= h;
			//kb5 = (Q*temp)*h;


			temp = b;
			temp.add(kb1,b61);
			temp.add(kb2,b62);
			temp.add(kb3,b63);
			temp.add(kb4,b64);
			temp.add(kb5,b65);
			//temp = kb1*b61+kb2*b62+kb3*b63+kb4*b64+kb5*b65+b;
			Q.rightmult(temp,kb6);
			kb6 *= h;
			//kb6 = (Q*temp)*h;

			vectr errb = kb1;
			errb *= dc1;
			errb.add(kb3,dc3);
			errb.add(kb4,dc4);
			errb.add(kb5,dc5);
			errb.add(kb6,dc6);
			//vectr errb = dc1*kb1+dc3*kb3+dc4*kb4+dc5*kb5+dc6*kb6;


			// check error:
			/*err = errb.absmax();
			err /= eps;
			if (err <= 1.0) break;*/
                        double bmin = b.absmin();
                        if(bmin<1.0) bmin = 1.0;
                        err = errb.absmax()/bmin;
                        err /= eps;
                        if(err<=1.0) break;

			// adjust step:
			double htemp = 0.9*h*pow(err,-0.25);
			if (htemp>h/10) h = htemp;
			else h /= 10;
			if (currt+h == currt) {
                            return ret;
                            /*cerr << "step size underflow in rk2" << endl;
				cerr << "in beta pass:" << endl;
				cerr << Q << endl;
				cerr << b << endl;
				cerr << t << ' ' << currt << endl;
				exit(1);*/
			}
		}
		currt += h;
		if (err > 1.89e-4) h = 0.9*h*pow(err,-0.2);
		else h *= 5;

		b.add(kb1,c1);
		b.add(kb3,c3);
		b.add(kb4,c4);
		b.add(kb6,c6);
		//b += c1*kb1+c3*kb3+c4*kb4+c6*kb6;
		double ss = b.sum();
		if (ss>0.0) {
			ret += log(ss);
			b /= ss;
		}
		//ret += log(b.normalize());
	}
#ifdef USECACHE
	cacheret[cachepos] = b;
	cachellh[cachepos] = ret;
	cachepos++;
	//if (cachepos>MAXCACHE) { cerr << "cache overflow" << endl; cachepos=0;}
	if (cachepos>=MAXCACHE) cachepos=0;
#endif
	return ret;
}


// calculates  \int_u=0^t a*exp(Q*u)*e_i e_j*exp(Q*(t-u))*b du
//       [this is a matrix indexed by (i,j)]
// the output is returned on c
// (actually, it returns the transpose...)
// This is a version based on adaptive step sizes (and therefore
// a fifth-order Runge-Kutta method with an embedded 4th order method
// -- see Numerical Recipes in C)
double suffstat(const vectr &a, const vectr &b, matrix &c,
              const matrix &Q, double t, double eps, double h) {
	double ret = 0.0;
	if (t<=0.0) return ret;

	static double defaulteps = ParamDouble("RKEPS",RKEPS);
	if (eps<0.0) eps = defaulteps;

#ifdef USECACHE
	static vector<vectr> cachea(MAXCACHE);
	static vector<vectr> cacheb(MAXCACHE);
	static vector<matrix> cacheQ(MAXCACHE);
	static vector<double> cachet(MAXCACHE);
	static vector<matrix> cacheret(MAXCACHE);
	static vector<double> cachellh(MAXCACHE);
	static int cachepos = 0;
	tries++;
	//if (tries%1000==0) cout << "cache: " << reuse << '/' << tries << endl;
	for(int i=0;i<cacheb.size();i++)
		if (cachea[i]==a && cacheb[i]==b && cacheQ[i]==Q && cachet[i]==t) {
			reuse++;
			c = cacheret[i];
			return cachellh[i];
		}
	cachea[cachepos] = a;
	cacheb[cachepos] = b;
	cacheQ[cachepos] = Q;
	cachet[cachepos] = t;
#endif

	int p = Q.getm();

	if (h<=0) {
		double dd = Q.diag().absmax();
		if (dd>0) h = 0.1/Q.diag().absmax();
		else h = 0.1;
	}

	double maxh = h, minh = h;
	//int evals = 0;

	vectr mya(a);
	c = matrix(p,p,0.0);
	vectr ka1(p),ka2(p),ka3(p),ka4(p),ka5(p),ka6(p);
	matrix kc1(p,p),kc2(p,p),kc3(p,p),kc4(p,p),kc5(p,p),kc6(p,p);

	//static double a2=0.2, a3=0.3, a4=0.6, a5=1.0, a6=0.875,
	static double
		b21=0.2,
		b31=3.0/40.0, b32=9.0/40.0,
		b41=0.3, b42=-0.9, b43=1.2,
		b51=-11.0/54.0, b52=2.5, b53=-70.0/27.0, b54=35.0/27.0,
		b61=1631.0/55296.0, b62=175.0/512.0, b63=575.0/13824.0,
		b64=44275.0/110592.0, b65=253.0/4096.0,
		c1=37.0/378.0, c3=250.0/621.0, c4=125.0/594.0, c6=512.0/1771.0,
		dc5=-277.00/14336.0;
	static double dc1=c1-2825.0/27648.0, dc3=c3-18575.0/48384.0,
		dc4=c4-13525.0/55296.0, dc6=c6-0.25;

	double currt = 0;
	while(currt<t) {
		if (h>maxh) maxh = h;
		if (h<minh) minh = h;
		if (currt+h>=t) h = t-currt; // don't overshoot!

		double err;

		for(;;) { // find stepsize
			//evals++;
			//if(evals>MAX_INT) return;
			// take 5th-order step:
			vectr temp;
			ka1 = (mya*Q)*h;
			kc1 = (matrix(b,mya)
					+ Q*c)*h;

			temp = ka1*b21+mya;
			ka2 = (temp*Q)*h;
			kc2 = (matrix(b,temp)
					+ Q*(kc1*b21+c))*h;

			temp = ka1*b31+ka2*b32+mya;
			ka3 = (temp*Q)*h;
			kc3 = (matrix(b,temp) 
				    	+ Q*(kc1*b31+kc2*b32+c))*h;

			temp = ka1*b41+ka2*b42+ka3*b43+mya;
			ka4 = (temp*Q)*h;
			kc4 = (matrix(b,temp)
					+ Q*(kc1*b41+kc2*b42+kc3*b43+c))*h;

			temp = ka1*b51+ka2*b52+ka3*b53+ka4*b54+mya;
			ka5 = (temp*Q)*h;
			kc5 = (matrix(b,temp)
					+ Q*(kc1*b51+kc2*b52+kc3*b53+kc4*b54+c))*h;

			temp = ka1*b61+ka2*b62+ka3*b63+ka4*b64+ka5*b65+mya;
			ka6 = (temp*Q)*h;
			kc6 = (matrix(b,temp)
					+ Q*(kc1*b61+kc2*b62+kc3*b63+kc4*b64+kc5*b65+c))*h;

			vectr erra = dc1*ka1+dc3*ka3+dc4*ka4+dc5*ka5+dc6*ka6;
			matrix errc = dc1*kc1+dc3*kc3+dc4*kc4+dc5*kc5+dc6*kc6;

			// check error:
			/*err = erra.absmax();
			double terr = errc.absmax();
			if (err<terr) err = terr;
			err /= eps;
			if (err <= 1.0) break;*/

                        double amin = a.absmin();
                        double bmin = b.absmin();
                        if(amin<1.0) amin = 1.0;
                        if(bmin<1.0) bmin = 1.0;
                        if(bmin>amin) amin = bmin;
                        err = erra.absmax()/amin;
                        double terr = errc.absmax()/amin;
                        if(err<terr) err = terr;
                        err /= eps;
                        if(err <= 1.0) break;

			// adjust step:
			double htemp = 0.9*h*pow(err,-0.25);
			if (htemp>h/10) h = htemp;
			else h /= 10;
			if (currt+h == currt) {
                            return ret;
                            /*cerr << "step size underflow in rk2" << endl;
				cerr << "in ss pass:" << endl;
				cerr << Q << endl;
				cerr << a << endl;
				cerr << b << endl;
				cerr << t << ' ' << currt << endl;
				exit(1);*/
			}
		}
		currt += h;
		if (err > 1.89e-4) h = 0.9*h*pow(err,-0.2);
		else h *= 5;

		mya += c1*ka1+c3*ka3+c4*ka4+c6*ka6;
		c += c1*kc1+c3*kc3+c4*kc4+c6*kc6;
		double ss = mya.sum();
		if (ss>0.0) {
			ret += log(ss);
			c /= ss;
			mya /= ss;
		}
		//double z=mya.normalize();
		//ret += log(z);
		//c/=z;
                
	}

#ifdef USECACHE
	cacheret[cachepos] = c;
	cachellh[cachepos] = ret;
	cachepos++;
	//if (cachepos>MAXCACHE) { cerr << "cache overflow" << endl; cachepos=0;}
	if (cachepos>=MAXCACHE) cachepos=0;
#endif
	return ret;
}
