/* Continuous Time Bayesian Network Reasoning and Learning Engine
 * Copyright (C) 2009 The Regents of the University of California
 *
 * Authored by Yu Fan, William Lam, Joon Lee, Christian Shelton, and Jing Xu
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "samplinginf.h"
#include "markov.h"
#include "markovdyn.h"
#include "multirv.h"
#include "ctbndyn.h"
#include "utils.h"

using namespace std;

SamplingInf::SamplingInf() {
	p = NULL;
	numofsamples = 100;
	timelimit = 0.0;
	isgibbs = false;
}

SamplingInf::~SamplingInf() {
	if(sampler!=NULL) delete sampler;
}

SamplingInf* SamplingInf::Clone() const {
	SamplingInf* siClone = new SamplingInf(*this);
	siClone->SetSampler(this->sampler->Clone());
	return siClone;
}

void SamplingInf::SetProcess(const Process *pr) {
	if(sampler==NULL) {
		cout << "Sampler not set!" << endl;
		exit(1);
	}
	sampler->SetProcess(pr);
	p = dynamic_cast<const Markov*>(pr);
	CleanSamples();
}

void SamplingInf::SetSampler(const Sampler *s) {
	sampler = s->Clone();
	if(p!=NULL) sampler->SetProcess(p);
}

void SamplingInf::SetTrajectory(const Trajectory *tr) {
	if(sampler==NULL) {
		cout << "Sampler not set!" << endl;
		exit(1);
	}
	if(p==NULL) {
		cout << "Process not set!" << endl;
		exit(1);
	}
	sampler->SetTrajectory(tr);
	CleanSamples();
}

void SamplingInf::SetNumSample(int n) {
	numofsamples = n;
}

void SamplingInf::SetTimeLimit(double t, bool gibbs) {
	timelimit = t;
	isgibbs = gibbs;
}

// Sample completions of given evidence trajectory (stored in sampler), and
// associate them with their weights.
void SamplingInf::Sample() {
	if(sampler==NULL || p==NULL) {
		cout << "Sampler or process not set!" << endl;
		exit(1);
	}
	if (timelimit == 0.0) {
		sampler->SampleTrajectories(traj, weights, numofsamples);
//		cerr << "+++++++ sampled trajectories +++++\n";
//		for (size_t i = 0; i < traj.size(); ++i)
//		traj[i].Draw(cerr);
//		cout << "weights: " << weights << endl;
	} else {
		int totalnum = 0;
		double begin_time = getcputime();
		double end_time = begin_time + timelimit;
		int kMinNumTraj= 10;
		if(isgibbs) kMinNumTraj= 2;
		do {
			vector<Trajectory> t;
			vector<double> w;
			sampler->SampleTrajectories(t, w, kMinNumTraj);
			for (int i = 0; i < kMinNumTraj; ++i) {
				traj.push_back(t[i]);
				weights.push_back(w[i]);
			}
			totalnum+=kMinNumTraj;
		} while (getcputime() < end_time);
		cout << "total samples: " << totalnum << endl;
		if (isgibbs) {
			vector<Trajectory>::iterator traj_iter = traj.begin();
			vector<double>::iterator weights_iter = 
							weights.begin();
			if(traj.size()<50) {
				// Discard the first half (as burn-in).
				traj.erase(traj_iter, 
						traj_iter + traj.size() / 2);
				weights.erase(weights_iter, 
						weights_iter 
						+ weights.size() / 2);
			}
			else {
				traj.erase(traj_iter, traj_iter + 50);
				weights.erase(weights_iter, 
						weights_iter + 50);

			}
		}
	}
	double maxweight = weights[0];
	for(int i=1; i<weights.size(); i++)
		if(weights[i]>maxweight) maxweight = weights[i];
	for(int i=0; i<weights.size(); i++)
		weights[i] = exp(weights[i] - maxweight);
}

void SamplingInf::CleanSamples() {
	traj.clear();
	weights.clear();
}

double SamplingInf::Smooth(const Instantiation &x, double t, bool log) {
	if(sampler==NULL) {
		cout << "smooth: Sampler not set!" << endl;
		exit(1);
	}
	if(traj.size()==0) Sample();
	//cout << "sample complete" << endl;
	Context querycontext = x.KnownVars();
	Instantiation queryvalue(querycontext, x);
	int queryindex = queryvalue.Index();
	double retprob = 0;
	double sumweights = 0;
	int count=1;
	for(int i=0; i<traj.size(); i++) {
		//traj[i].Draw(cout);
		Instantiation sampleval = traj[i].Values(querycontext, t);
		//cout << sampleval << endl;
		if(sampleval.Index()==queryindex) {
			retprob += weights[i];
			count++;
			//cout << count << endl;
		}
		sumweights += weights[i];
	}
	//cout << "count: " << retprob << ", weight: " << sumweights << endl;
	retprob /= sumweights;
	if(log)
		return ::log(retprob);
	else
		return retprob;
}

double SamplingInf::Filter(const Instantiation &x, double t, bool log) {
	cout << "not implemented" << endl;
	return 0;
}


double SamplingInf::Prob(double t, bool log) {
	cout << "not implemented" << endl;
	return 0;
}

void SamplingInf::AddExpSuffStats(SS *ss, double w) {
	if(sampler==NULL) {
		cout << "Sampler not set!" << endl;
		exit(1);
	}

	if(traj.size()==0) Sample();
	double sumweights = 0.0;
	for(int i=0; i<traj.size(); i++) {
		if(isnan(weights[i])) {
			cerr << "Suff Stats NaN:" << endl;
			//traj[i].Draw(cout);
			cerr << endl;
			cerr << weights << endl;
			exit(0);
		}
		sumweights += weights[i];
	}

	SS *newss = p->SuffStats(traj, weights);
	newss->Scale(sumweights); //by Yu
	ss->AddSS(newss);
	delete newss;
}

void SamplingInf::AddExpSuffStats(const Dynamics *dyn, SS *ss,
							  double w) {
	if(sampler==NULL) {
		cout << "Sampler not set!" << endl;
		exit(1);
	}
	if(traj.size()==0) Sample();
	double sumweights = 0.0;
	for(int i=0; i<traj.size(); i++) {
		if(isnan(weights[i])) {
			cerr << "Suff Stats NaN:" << endl;
			//traj[i].Draw(cout);
			cerr << endl;
			cerr << weights << endl;
			exit(0);
		}
		sumweights += weights[i];
	}
	SS *newdynss = dyn->SuffStats(traj, weights);
	newdynss->Scale(sumweights);
	ss->AddSS(newdynss);
	delete newdynss;
}

void SamplingInf::AddExpSuffStats(const RV *p0, SS *ss,
						  double w) {
	if(sampler==NULL) {
		cout << "Sampler not set!" << endl;
		exit(1);
	}
	if(traj.size()==0) Sample();
	double sumweights = 0.0;
	for(int i=0; i<traj.size(); i++) {
		if(isnan(weights[i])) {
			cerr << "Suff Stats NaN:" << endl;
			//traj[i].Draw(cout);
			cerr << endl;
			cerr << weights << endl;
			exit(0);
		}
		sumweights += weights[i];
	}
	SS *newp0ss = p0->SuffStats(traj, weights);
	newp0ss->Scale(sumweights);
	ss->AddSS(newp0ss);
	delete newp0ss;
}

double SamplingInf::CalcQuery(QueryCalculator &calc) {
	if(sampler==NULL) {
		cout << "Sampler not set!" << endl;
		exit(1);
	}

	if(traj.size()==0) Sample();

	double sumweights = 0;
	for(int i=0; i<traj.size(); i++)
		sumweights += weights[i];
	//   for(int i=0; i<traj.size(); i++)
	//     weights[i] /= sumweights;

	double retval = 0.0;

	int n = traj.size();
	for(int i=0; i<n; i++)
		retval += weights[i] * calc.Calculate(traj[i]);
	return retval/sumweights;
}

