/* Continuous Time Bayesian Network Reasoning and Learning Engine
 * Copyright (C) 2009 The Regents of the University of California
 *
 * Authored by Yu Fan, William Lam, Joon Lee, Christian Shelton, and Jing Xu
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
// this file reads in a model file and an evidence file, do the query using gibbs sampling
#include "markovdyn.h"
#include "streamextra.h"
#include "context.h"
#include "multirv.h"
#include "samplinginf.h"
#include "gibbssampler.h"
#include "params.h"
#include "bn.h"
#include "ctbndyn.h"
#include "em.h"
#include "utils.h"
#include <iostream>

using namespace std;

namespace {
	int testing_result = 0;
	double kEpsilon = 0.15;
};

#define EXPECT(cond) do { \
        if (!(cond)) {                                                  \
			std::cerr << "Test failure at " << __FILE__ << ":" << __LINE__ \
                      << " in " << __func__ \
                      << ", failed condition is " #cond << std::endl; \
            testing_result = 1; \
        } \
    } while (0)

int main(int argc, char **argv)
{
	randomizer.Reset(1);
	const char *input_file = "queryinput.data";
	addnan(cout);
	InitParams(argc, argv);
 
	ifstream fin(input_file);
	Markov m(fin);
	const CTBNDyn *ctbndyn = dynamic_cast<const CTBNDyn *>(m.GetDynamics());
	const BN *bn = dynamic_cast<const BN *>(m.GetStartDist());
	Context context = ctbndyn->Domain() + ctbndyn->CondDomain();
	Trajectory evid;
	fin >> evid;
	double expect_q1, expect_q2;
	fin >> expect_q1 >> expect_q2;
	fin.close();

	Instantiation x(context, -1);
	int querynode = ParamInt("QueryNode", 1);
	int querytimeval = ParamInt("QueryTimeVal", 1);
	Context c = ctbndyn->Node(querynode)->Domain();
	x.SetVal(querynode, querytimeval);
	QueryTime query1(x);
	QueryTransition query2(c, 0, 1); 

	SamplingInf inf;
	int numofburniniter = 1000;
	int numofsamples = 1000;
	Sampler *sampler;
	sampler = new GibbsSampler(&m, &evid, numofburniniter);
	inf.SetSampler(sampler);
	inf.SetProcess(&m);
	inf.SetNumSample(numofsamples);
	double error1 = (inf.CalcQuery(query1) - expect_q1) / expect_q1;
	double error2 = (inf.CalcQuery(query2) - expect_q2) / expect_q2;
	EXPECT(fabs(error1) < kEpsilon);
	EXPECT(fabs(error2) < kEpsilon);
	if (testing_result == 0)
		cout << "PASS gibbsquery unittest." << endl;
	return 0;
}
