/* Continuous Time Bayesian Network Reasoning and Learning Engine
 * Copyright (C) 2009 The Regents of the University of California
 *
 * Authored by Yu Fan, William Lam, Joon Lee, Christian Shelton, and Jing Xu
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "em.h"
#include "params.h"
#include "utils.h"
#include "markov.h"
#include "exactmarkovinf.h"
#include "structure.h"
#include "ctbndyn.h"
#include "dynamics.h"
#include "expsuffstatsquery.h"
#include "nonexpsuffstatsquery.h"
#include "brutestructuresearch.h"
#include "grapheditsearch.h"
#include "famscore.h"

using namespace std;

void EM(const vector<Trajectory> &data, Process *dist, Inference *inf) {
	int nitt = ParamInt("EMItt", 5); 
	int ndata = data.size();

	for(int i = 0; i < nitt; i++) {
		inf->SetProcess(dist);
		SS *ss = dist->BlankSS();
		for(int j=0;j<ndata;j++) {
			inf->SetTrajectory(&data[j]);
			inf->AddExpSuffStats(ss);
		}
		dist->Maximize(ss);
		cerr << endl;
		delete ss;
	}
}

void SEM(const vector<Trajectory> &data, 
		Process *&dist, 
		Inference *inf) {
	int nitt = ParamInt("SEMItt", 10);
	double nTrans = ParamDouble("ptrans", 1.0);
	double nTime = ParamDouble("ptime", 1.0);
	Structure s;

	// Initial parameter update
	EM(data,dist,inf);

	for(int i = 0; i < nitt; i++) {
		ExpSuffStatsQuery* essQuery = 
			new ExpSuffStatsQuery(inf->Clone(),true);
		essQuery->SetData(&data);
		essQuery->SetProcess(dist);

		// Structure modification step
		Markov* m = dynamic_cast<Markov *>(dist);
		const BN* bnet = dynamic_cast<const BN *>(m->GetStartDist());
		
		const CTBNDyn* cdyn = 
			dynamic_cast<const CTBNDyn *>(m->GetDynamics());
		
		BNFamScore* bfs = new BNFamScore(nTrans,essQuery);
		GraphEditSearch gesBN(bnet->Domain(),bfs,true);
		s = gesBN.LearnStructure();
		delete bfs;
		
		RV* newrv = new BN(s, bnet->Domain());
		dynamic_cast<BN *>(newrv)->FillParams(*bnet);
		
		CTBNFamScore* cfs = new CTBNFamScore(nTrans,nTime,essQuery);
		GraphEditSearch ges(cdyn->Domain(),cfs);
		s = ges.LearnStructure();
		delete cfs;

		delete essQuery;

		Dynamics* newdyn = new CTBNDyn(s, cdyn->Domain());
		dynamic_cast<CTBNDyn *>(newdyn)->FillParams(*cdyn);

		Process* newdist = new Markov(newrv, newdyn);
		delete dist;
		dist = newdist;
		
		// Inference and parameter update step
		EM(data, dist, inf);
	}
}

